#!/usr/bin/python
# backup_control.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_control.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: backup_control.

A high level functions to manage backups. Keeps track of current
``Jobs`` and ``Tasks``. The "Jobs" dictionary keeps already started
backups ( by backupID ) objects, see ``p2p.backup`` module. "Tasks" is a
list of path IDs to start backups in the future, as soon as some "Jobs"
gets finished.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from io import StringIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys
import time
import zlib
import pprint

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor backup_control.py')

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import tmpfile
from bitdust.system import dirsize

from bitdust.lib import misc
from bitdust.lib import packetid
from bitdust.lib import nameurl
from bitdust.lib import strng

from bitdust.main import settings
from bitdust.main import events

from bitdust.raid import eccmap

from bitdust.crypt import encrypted
from bitdust.crypt import key
from bitdust.crypt import my_keys

from bitdust.userid import global_id
from bitdust.userid import id_url

from bitdust.services import driver

from bitdust.contacts import contactsdb

from bitdust.p2p import p2p_service

from bitdust.stream import data_sender

from bitdust.storage import backup_fs
from bitdust.storage import backup_matrix
from bitdust.storage import backup

from bitdust.userid import my_id

#------------------------------------------------------------------------------

MAXIMUM_JOBS_STARTED = 1  # let's do only one backup at once for now

_Jobs = {}  # here are already started backups ( by backupID )
_Tasks = []  # here are tasks to start backups in the future ( pathID )
_LastTaskNumber = 0
_LoadingFlag = False
_TaskStartedCallbacks = {}
_TaskFinishedCallbacks = {}

#------------------------------------------------------------------------------


def jobs():
    """
    Mutator method to access ``Jobs`` dictionary.
    """
    global _Jobs
    return _Jobs


def tasks():
    """
    Mutator method to access ``Tasks`` list.
    """
    global _Tasks
    return _Tasks


#------------------------------------------------------------------------------


def init():
    """
    Must be called before other methods here.

    Load index database from file `~/.bitdust/[network name]/metadata/index`.
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.init')


def shutdown():
    """
    Called for the correct completion of all things.
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.shutdown')


#------------------------------------------------------------------------------


def SaveFSIndex(customer_idurl=None, key_alias='master', increase_revision=True):
    """
    Save index data base to local file and notify "index_synchronizer()" or "shared_access_coordinator()" state machines.
    """
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, k=key_alias, increase_revision=increase_revision)
    global _LoadingFlag
    if _LoadingFlag:
        return False
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if increase_revision:
        backup_fs.commit(
            customer_idurl=customer_idurl,
            key_alias=key_alias,
        )
    backup_fs.SaveIndex(customer_idurl, key_alias)
    if increase_revision:
        if key_alias == 'master':
            if driver.is_on('service_backup_db'):
                from bitdust.storage import index_synchronizer
                index_synchronizer.A('push')
        else:
            if driver.is_on('service_shared_data'):
                from bitdust.access import shared_access_coordinator
                shared_access_coordinator.on_index_file_updated(customer_idurl, key_alias)


#------------------------------------------------------------------------------


def on_files_received(newpacket, info):
    list_files_global_id = global_id.ParseGlobalID(newpacket.PacketID)
    if not list_files_global_id['idurl']:
        lg.warn('invalid PacketID: %s' % newpacket.PacketID)
        return False
    if list_files_global_id['key_alias'] != 'master':
        # ignore Files() if this is a shared data
        if _Debug:
            lg.dbg(_DebugLevel, 'ignore incoming %r, this is not a private data' % newpacket)
        return False
    if not id_url.is_the_same(list_files_global_id['idurl'], my_id.getIDURL()):
        # ignore Files() if this is another customer
        if _Debug:
            lg.dbg(_DebugLevel, 'ignore incoming %r which is owned by another customer' % newpacket)
        return False
    if not contactsdb.is_supplier(newpacket.OwnerID):
        # ignore Files() if this is not my supplier
        if _Debug:
            lg.dbg(_DebugLevel, 'incoming %r received, but %r is not my supplier' % (newpacket, newpacket.OwnerID))
        return False
    if _Debug:
        lg.dbg(_DebugLevel, '%r for us from %s' % (newpacket, newpacket.CreatorID))
    if IncomingSupplierListFiles(newpacket, list_files_global_id):
        p2p_service.SendAck(newpacket)
    else:
        p2p_service.SendFail(newpacket)
    return True


#------------------------------------------------------------------------------


def UnpackListFiles(payload, method):
    if method == 'Text':
        return payload
    elif method == 'Compressed':
        return strng.to_text(zlib.decompress(strng.to_bin(payload)))
    return payload


def IncomingSupplierListFiles(newpacket, list_files_global_id):
    """
    Called when command "Files" were received from one of my suppliers.
    This is an answer from given supplier (after my request) to get a
    list of our files stored on his machine.
    """
    supplier_idurl = newpacket.OwnerID
    customer_idurl = list_files_global_id['idurl']
    num = contactsdb.supplier_position(supplier_idurl, customer_idurl=customer_idurl)
    if num < -1:
        lg.warn('unknown supplier: %s' % supplier_idurl)
        return False
    from bitdust.customer import list_files_orator
    target_key_id = my_keys.latest_key_id(list_files_global_id['key_id'])
    if not my_keys.is_key_private(target_key_id):
        lg.warn('key %r not registered, not possible to decrypt ListFiles() packet from %r' % (target_key_id, supplier_idurl))
        return False
    try:
        block = encrypted.Unserialize(
            newpacket.Payload,
            decrypt_key=target_key_id,
        )
        input_data = block.Data()
    except:
        lg.err('failed decrypting data from packet %r received from %r' % (newpacket, supplier_idurl))
        return False
    from bitdust.storage import index_synchronizer
    is_in_sync = index_synchronizer.is_synchronized() and backup_fs.revision() > 0
    list_files_raw = UnpackListFiles(input_data, settings.ListFilesFormat())
    remote_files_changed, backups2remove, paths2remove, missed_backups = backup_matrix.process_raw_list_files(
        supplier_num=num,
        list_files_text_body=list_files_raw,
        customer_idurl=None,
        is_in_sync=is_in_sync,
    )
    list_files_orator.IncomingListFiles(newpacket)
    if remote_files_changed:
        backup_matrix.SaveLatestRawListFiles(supplier_idurl, list_files_raw)
    if _Debug:
        lg.args(_DebugLevel, s=nameurl.GetName(supplier_idurl), c=nameurl.GetName(customer_idurl), backups2remove=len(backups2remove), paths2remove=len(paths2remove), files_changed=remote_files_changed, missed_backups=len(missed_backups))
    if len(backups2remove) > 0:
        p2p_service.RequestDeleteListBackups(backups2remove)
        backup_matrix.populate_remote_versions_deleted(backups2remove)
        if _Debug:
            lg.out(_DebugLevel, '    also sent requests to remove %d backups' % len(backups2remove))
    if len(paths2remove) > 0:
        p2p_service.RequestDeleteListPaths(paths2remove)
        if _Debug:
            lg.out(_DebugLevel, '    also sent requests to remove %d paths' % len(paths2remove))
    if len(missed_backups) > 0:
        from bitdust.storage import backup_rebuilder
        backup_rebuilder.AddBackupsToWork(missed_backups)
        backup_rebuilder.A('start')
        if _Debug:
            lg.out(_DebugLevel, '    also triggered service_rebuilding with %d missed backups' % len(missed_backups))
    del backups2remove
    del paths2remove
    del missed_backups
    return True


def IncomingSupplierBackupIndex(newpacket, key_id=None, deleted_path_ids=[]):
    """
    Called by ``p2p.p2p_service`` when a remote copy of our local index data
    base ( in the "Data" packet ) is received from one of our suppliers.

    The index is also stored on suppliers to be able to restore it.
    """
    block = encrypted.Unserialize(newpacket.Payload, decrypt_key=key_id)
    if not block:
        lg.err('failed reading data from %s' % newpacket.RemoteID)
        return None
    try:
        data = block.Data()
        inpt = StringIO(strng.to_text(data))
        supplier_revision = inpt.readline().rstrip('\n')
        if supplier_revision:
            supplier_revision = int(supplier_revision)
        else:
            supplier_revision = None
        text_data = inpt.read()
    except:
        lg.exc()
        try:
            inpt.close()
        except:
            pass
        return None
    inpt.close()
    if _Debug:
        lg.args(_DebugLevel, k=key_id, p=newpacket.PacketID, c=newpacket.CreatorID, sz=len(text_data), inp=len(newpacket.Payload), deleted=len(deleted_path_ids))
    count, updated_customers_keys = backup_fs.ReadIndex(text_data, new_revision=supplier_revision, deleted_path_ids=deleted_path_ids)
    if updated_customers_keys:
        for customer_idurl, key_alias in updated_customers_keys:
            backup_fs.SaveIndex(customer_idurl, key_alias)
            if _Debug:
                lg.out(_DebugLevel, 'backup_control.IncomingSupplierBackupIndex updated to revision %d for %s of %s from %s' % (backup_fs.revision(customer_idurl, key_alias), customer_idurl, key_alias, newpacket.RemoteID))
    return supplier_revision


#------------------------------------------------------------------------------


def DeleteAllBackups():
    """
    Remove all backup IDs from index data base, see ``DeleteBackup()`` method.
    """
    # prepare a list of all known backup IDs
    all_ids = set(backup_fs.ListAllBackupIDs(customer_idurl=my_id.getIDURL()))
    all_ids.update(backup_matrix.GetBackupIDs(remote=True, local=True))
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.DeleteAllBackups %d ID\'s to kill' % len(all_ids))
    # delete one by one
    for backupID in all_ids:
        DeleteBackup(backupID, saveDB=False, calculate=False)
    # scan all files
    backup_fs.Scan()
    # check and calculate used space
    backup_fs.Calculate()
    # save the index
    SaveFSIndex()


def DeleteBackup(backupID, removeLocalFilesToo=True, saveDB=True, calculate=True):
    """
    This removes a single backup ID completely. Perform several operations:

    1) abort backup if it just started and is running at the moment
    2) if we requested for files for this backup we do not need it anymore - remove 'Data' requests
    3) remove interests in transport_control, see ``lib.transport_control.DeleteBackupInterest()``
    4) remove that ID from the index data base
    5) remove local files for this backup ID
    6) remove all remote info for this backup from the memory, see ``p2p.backup_matrix.EraseBackupRemoteInfo()``
    7) also remove local info from memory, see ``p2p.backup_matrix.EraseBackupLocalInfo()``
    8) stop any rebuilding, we will restart it soon
    9) check and calculate used space
    10) save the modified index data base, soon it will be synchronized with "index_synchronizer()" state machine
    """
    backupID = global_id.CanonicalID(backupID)
    key_alias, customer, _, _ = packetid.SplitBackupIDFull(backupID)
    customer_idurl = global_id.GlobalUserToIDURL(customer)
    # if the user deletes a backup, make sure we remove any work we're doing on it
    # abort backup if it just started and is running at the moment
    if AbortRunningBackup(backupID):
        if _Debug:
            lg.out(_DebugLevel, 'backup_control.DeleteBackup %s is in process, stopping' % backupID)
        return True
    from bitdust.stream import io_throttle
    from bitdust.storage import backup_rebuilder
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.DeleteBackup %r' % backupID)
    # if we requested for files for this backup - we do not need it anymore
    io_throttle.DeleteBackupRequests(backupID)
    io_throttle.DeleteBackupSendings(backupID)
    # remove interests in transport_control
    # mark it as being deleted in the db, well... just remove it from the index now
    if not backup_fs.DeleteBackupID(backupID):
        return False
    # finally remove local files for this backupID
    if removeLocalFilesToo:
        backup_fs.DeleteLocalBackup(settings.getLocalBackupsDir(), backupID)
    # remove all remote info for this backup from the memory
    backup_matrix.EraseBackupRemoteInfo(backupID)
    # also remove local info
    backup_matrix.EraseBackupLocalInfo(backupID)
    # stop any rebuilding, we will restart it soon
    backup_rebuilder.RemoveAllBackupsToWork()
    backup_rebuilder.SetStoppedFlag()
    # check and calculate used space
    if calculate or key_alias != 'master':
        backup_fs.Scan(customer_idurl=customer_idurl, key_alias=key_alias)
        backup_fs.Calculate(iterID=backup_fs.fsID(customer_idurl, key_alias))
    # in some cases we want to save the DB later
    if saveDB or key_alias != 'master':
        SaveFSIndex(customer_idurl, key_alias)
    return True


def DeletePathBackups(pathID, removeLocalFilesToo=True, saveDB=True, calculate=True):
    """
    This removes all backups of given path ID
    Doing same operations as ``DeleteBackup()``.
    """
    from bitdust.storage import backup_rebuilder
    from bitdust.stream import io_throttle
    pathID = global_id.CanonicalID(pathID)
    # get the working item
    customer, remotePath = packetid.SplitPacketID(pathID)
    key_alias = packetid.KeyAlias(customer)
    customer_idurl = global_id.GlobalUserToIDURL(customer)
    item = backup_fs.GetByID(remotePath, iterID=backup_fs.fsID(customer_idurl, key_alias))
    if item is None:
        return False
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.DeletePathBackups ' + pathID)
    # this is a list of all known backups of this path
    versions = item.list_versions()
    for version in versions:
        backupID = packetid.MakeBackupID(customer, remotePath, version, key_alias=key_alias)
        if _Debug:
            lg.out(_DebugLevel, '        removing %s' % backupID)
        # abort backup if it just started and is running at the moment
        AbortRunningBackup(backupID)
        # if we requested for files for this backup - we do not need it anymore
        io_throttle.DeleteBackupRequests(backupID)
        io_throttle.DeleteBackupSendings(backupID)
        # remove interests in transport_control
        # remove local files for this backupID
        if removeLocalFilesToo:
            backup_fs.DeleteLocalBackup(settings.getLocalBackupsDir(), backupID)
        # remove remote info for this backup from the memory
        backup_matrix.EraseBackupRemoteInfo(backupID)
        # also remove local info
        backup_matrix.EraseBackupLocalInfo(backupID)
        # finally remove this backup from the index
        item.delete_version(version)
        # lg.out(8, 'backup_control.DeletePathBackups ' + backupID)
    # stop any rebuilding, we will restart it soon
    backup_rebuilder.RemoveAllBackupsToWork()
    backup_rebuilder.SetStoppedFlag()
    # check and calculate used space
    if calculate or key_alias != 'master':
        backup_fs.Scan(customer_idurl=customer_idurl, key_alias=key_alias)
        backup_fs.Calculate(iterID=backup_fs.fsID(customer_idurl, key_alias))
    # save the index if needed
    if saveDB or key_alias != 'master':
        SaveFSIndex(customer_idurl, key_alias)
    return True


#------------------------------------------------------------------------------


def NewTaskNumber():
    """
    A method to create a unique number for new task.

    It just increments a variable in memory and returns it.
    """
    global _LastTaskNumber
    _LastTaskNumber += 1
    return _LastTaskNumber


class Task():

    """
    A class to represent a ``Task`` - a path to be backed up as soon as other backups will be finished.
    All tasks are stored in the list, see ``tasks()`` method.
    As soon as task get executed it fires the result call back and removed from the list.
    When task executes a new backup job gets created.
    """

    def __init__(self, pathID, localPath=None, keyID=None):
        self.number = NewTaskNumber()  # index number for the task
        self.created = time.time()
        self.backupID = None
        self.pathID = None
        self.dataID = None
        self.totalSize = None
        self.fullGlobPath = None
        self.fullCustomerID = None
        self.customerGlobID = None
        self.customerIDURL = None
        self.remotePath = None
        self.localPath = None
        self.sourcePath = None
        self.keyID = None
        self.keyAlias = None
        self.result_defer = Deferred()
        self.result_defer.addCallback(OnTaskExecutedCallback)
        self.result_defer.addErrback(OnTaskFailedCallback)
        parts = self.set_path_id(pathID)
        self.set_key_id(keyID or my_keys.make_key_id(alias=parts['key_alias'], creator_glob_id=parts['customer']))
        self.set_local_path(localPath)
        if _Debug:
            lg.out(_DebugLevel, 'new Task created: %r' % self)
        data = dict(
            number=self.number,
            created=self.created,
            backup_id=self.backupID,
            key_id=self.keyID,
            path_id=self.pathID,
            customer_id=self.customerGlobID,
            path=self.remotePath,
            local_path=self.localPath,
            remote_path=self.fullGlobPath,
        )
        events.send('backup-task-created', data=data)

    def __repr__(self):
        """
        Return a string like:

            "Task-5: 0/1/2/3 from /home/veselin/Documents/myfile.txt"
        """
        return 'Task-%d(%s from %s)' % (self.number, self.pathID, self.localPath)

    def destroy(self, message=None):
        if _Debug:
            lg.out(_DebugLevel, 'backup_control.Task-%d.destroy %s -> %s' % (self.number, self.localPath, self.backupID))
        if self.result_defer and not self.result_defer.called:
            self.result_defer.cancel()
            self.result_defer = None
        # yapf: disable
        events.send('backup-task-finished', data=dict(
            number=self.number,
            created=self.created,
            backup_id=self.backupID,
            key_id=self.keyID,
            path_id=self.pathID,
            customer_id=self.customerGlobID,
            path=self.remotePath,
            local_path=self.localPath,
            remote_path=self.fullGlobPath,
            message=message,
        ))
        # yapf: enable

    def set_path_id(self, pathID):
        parts = global_id.NormalizeGlobalID(pathID)
        self.pathID = pathID  # source path to backup
        self.customerGlobID = parts['customer']
        self.customerIDURL = parts['idurl']
        self.remotePath = parts['path']  # here it must be in 0/1/2 form
        if parts['key_alias']:
            self.set_key_id(my_keys.make_key_id(alias=parts['key_alias'], creator_glob_id=self.customerGlobID))
        return parts

    def set_key_id(self, key_id):
        self.keyID = key_id
        self.keyAlias = packetid.KeyAlias(self.keyID)
        self.fullGlobPath = global_id.MakeGlobalID(customer=self.customerGlobID, key_alias=self.keyAlias, path=self.remotePath)
        self.fullCustomerID = global_id.MakeGlobalID(customer=self.customerGlobID, key_alias=self.keyAlias)

    def set_local_path(self, localPath):
        self.localPath = localPath

    def run(self):
        """
        Runs a new ``Job`` from that ``Task``.
        """
        iter_and_path = backup_fs.WalkByID(self.remotePath, iterID=backup_fs.fsID(self.customerIDURL, self.keyAlias))
        if iter_and_path is None:
            if _Debug:
                lg.out(_DebugLevel, 'backup_control.Task.run ERROR %s not found in the index' % self.remotePath)
            err = 'remote path "%s" not found in the catalog' % self.remotePath
            OnTaskFailed(self.pathID, err)
            return err
        itemInfo, sourcePath = iter_and_path
        self.sourcePath = sourcePath
        if isinstance(itemInfo, dict):
            try:
                itemInfo = itemInfo[backup_fs.INFO_KEY]
            except:
                lg.exc()
                err = 'catalog item related to "%s" is broken' % self.remotePath
                OnTaskFailed(self.pathID, err)
                return err
        if not self.localPath:
            self.localPath = sourcePath
            lg.out('backup_control.Task.run local path was populated from catalog: %s' % self.localPath)
        if self.localPath != sourcePath:
            if _Debug:
                lg.out(_DebugLevel, '    local path different in the catalog: %s != %s' % (self.localPath, sourcePath))
        if not bpio.pathExist(self.localPath):
            lg.warn('path not exist: %s' % self.localPath)
            err = 'local path "%s" not exist' % self.localPath
            OnTaskFailed(self.pathID, err)
            return err
        dataID = misc.NewBackupID()
        if itemInfo.has_version(dataID):
            # ups - we already have same version
            # let's add 1,2,3... to the end to make absolutely unique version ID
            i = 1
            while itemInfo.has_version(dataID + str(i)):
                i += 1
            dataID += str(i)
        self.dataID = dataID
        self.backupID = packetid.MakeBackupID(
            customer=self.fullCustomerID,
            path_id=self.remotePath,
            version=self.dataID,
        )
        if self.backupID in jobs():
            lg.warn('backup job %s already started' % self.backupID)
            return 'backup job %s already started' % self.backupID
        if itemInfo.type == backup_fs.DIR:
            dirsize.ask(self.localPath, self.on_folder_size_counted, itemInfo)
        else:
            sz = os.path.getsize(self.localPath)
            reactor.callLater(0, self.on_folder_size_counted, self.localPath, sz, itemInfo)  # @UndefinedVariable
        return None

    def on_folder_size_counted(self, pth, sz, itemInfo):
        """
        This is a callback, fired from ``lib.dirsize.ask()`` method after finish
        calculating of folder size.
        """
        self.totalSize = sz
        if id_url.is_the_same(self.customerIDURL, my_id.getIDURL()):
            # TODO: need to rethink that approach
            # here not taking in account compressing rate of the local files
            # but taking in account remote version size - it is always doubled
            if int(backup_fs.sizebackups()) + self.totalSize*2 > settings.getNeededBytes():
                err = 'insufficient storage space expected'
                pth_id = self.pathID
                self.result_defer.errback((pth_id, err))
                reactor.callLater(0, OnTaskFailed, pth_id, err)  # @UndefinedVariable
                self.destroy(err)
                return None
        try:
            backup_fs.MakeLocalDir(settings.getLocalBackupsDir(), self.backupID)
        except:
            lg.exc()
            if _Debug:
                lg.out(_DebugLevel, 'backup_control.Task.on_folder_size_counted ERROR creating destination folder for %s' % self.pathID)
            err = 'failed creating destination folder for "%s"' % self.backupID
            pth_id = self.pathID
            self.result_defer.errback((pth_id, err))
            reactor.callLater(0, OnTaskFailed, pth_id, err)  # @UndefinedVariable
            self.destroy(err)
            return None

        itemInfo.set_size(self.totalSize)
        itemInfo.add_version(self.dataID)

        if _Debug:
            lg.out(_DebugLevel, 'backup_control.Task.on_folder_size_counted %s %d for %r' % (pth, sz, itemInfo))

        compress_mode = 'bz2'
        arcname = os.path.basename(self.sourcePath)

        from bitdust.storage import backup_tar
        if bpio.pathIsDir(self.localPath):
            backupPipe = backup_tar.backuptardir_thread(self.localPath, arcname=arcname, compress=compress_mode)
        else:
            backupPipe = backup_tar.backuptarfile_thread(self.localPath, arcname=arcname, compress=compress_mode)

        job = backup.backup(
            self.backupID,
            backupPipe,
            finishCallback=OnJobDone,
            blockResultCallback=OnBackupBlockReport,
            notifyNewDataCallback=OnNewDataPrepared,
            blockSize=settings.getBackupBlockSize(),
            sourcePath=self.localPath,
            keyID=self.keyID or itemInfo.key_id,
        )
        job.totalSize = self.totalSize
        jobs()[self.backupID] = job

        backup_fs.Calculate(iterID=backup_fs.fsID(self.customerIDURL, self.keyAlias))
        SaveFSIndex(customer_idurl=self.customerIDURL, key_alias=self.keyAlias)

        jobs()[self.backupID].automat('start')
        reactor.callLater(0, FireTaskStartedCallbacks, self.pathID, self.dataID)  # @UndefinedVariable

        if self.keyAlias == 'master':
            if driver.is_on('service_backup_db'):
                from bitdust.storage import index_synchronizer
                index_synchronizer.A('push')

        if _Debug:
            lg.out(_DebugLevel, 'backup_control.Task-%d.run [%s/%s], size=%d, %s' % (self.number, self.pathID, self.dataID, itemInfo.size, self.localPath))

        self.result_defer.callback((self.backupID, None))
        self.destroy(None)
        return None


#------------------------------------------------------------------------------


def PutTask(pathID, localPath=None, keyID=None):
    """
    Creates a new backup ``Task`` and append it to the list of tasks.
    """
    pathID = global_id.CanonicalID(pathID)
    current_task = GetPendingTask(pathID)
    if current_task:
        current_task.set_path_id(pathID)
        if localPath:
            current_task.set_local_path(localPath)
        if keyID:
            current_task.set_key_id(keyID)
        return current_task
    t = Task(pathID=pathID, localPath=localPath, keyID=keyID)
    tasks().append(t)
    return t


def DeleteAllTasks():
    """
    Clear the tasks list.
    """
    global _Tasks
    _Tasks = []


def RunTask():
    """
    Checks current jobs and run a one task if it is possible.
    Verifies if it is possible to start a new task,
    the maximum number of simultaneously running ``Jobs`` is limited.
    """
    if len(tasks()) == 0:
        return False
    if len(jobs()) >= MAXIMUM_JOBS_STARTED:
        return False
    from bitdust.customer import fire_hire
    if not fire_hire.IsAllHired():
        if _Debug:
            lg.out(_DebugLevel, 'backup_control.RunTask some suppliers not hired yet, retry after 5 sec')
        reactor.callLater(5, RunTask)  # @UndefinedVariable
        return False
    T = tasks().pop(0)
    message = T.run()
    if message:
        events.send('backup-task-failed', data=dict(
            path_id=T.pathID,
            message=message,
        ))
        T.result_defer.errback((T.pathID, message))
        T.destroy(message)
    return True


def HasTask(pathID):
    """
    Looks for path ID in the tasks list.
    """
    pathID = global_id.CanonicalID(pathID)
    for task in tasks():
        if task.pathID == pathID:
            return True
    return False


def GetPendingTask(pathID):
    pathID = global_id.CanonicalID(pathID)
    for t in tasks():
        if t.pathID == pathID:
            return t
    return None


def ListPendingTasks():
    return tasks()


def AbortPendingTask(pathID):
    pathID = global_id.CanonicalID(pathID)
    for t in tasks():
        if t.pathID == pathID:
            tasks().remove(t)
            return True
    return False


#------------------------------------------------------------------------------


def OnFoundFolderSize(pth, sz, arg):
    """
    This is a callback, fired from ``lib.dirsize.ask()`` method after finish
    calculating of folder size.
    """
    try:
        pathID, version = arg
        customerGlobID, pathID = packetid.SplitPacketID(pathID)
        customerIDURL = global_id.GlobalUserToIDURL(customerGlobID)
        keyAlias = packetid.KeyAlias(customerGlobID)
        item = backup_fs.GetByID(pathID, iterID=backup_fs.fsID(customerIDURL, keyAlias))
        if item:
            item.set_size(sz)
            backup_fs.Calculate(iterID=backup_fs.fsID(customerIDURL, keyAlias))
            SaveFSIndex(customerIDURL, keyAlias)
        if version:
            backupID = packetid.MakeBackupID(customerGlobID, pathID, version, key_alias=keyAlias)
            job = GetRunningBackupObject(backupID)
            if job:
                job.totalSize = sz
        if _Debug:
            lg.out(_DebugLevel, 'backup_control.OnFoundFolderSize %s %d' % (backupID, sz))
    except:
        lg.exc()


def OnJobDone(backupID, result):
    """
    A callback method fired when backup is finished.

    Here we need to save the index data base.
    """
    from bitdust.storage import backup_rebuilder
    lg.info('job done [%s] with result "%s", %d more tasks' % (backupID, result, len(tasks())))
    jobs().pop(backupID)
    keyAlias, customerGlobalID, remotePath, version = packetid.SplitBackupIDFull(backupID)
    customer_idurl = global_id.GlobalUserToIDURL(customerGlobalID)
    if result == 'done':
        maxBackupsNum = settings.getBackupsMaxCopies()
        if maxBackupsNum:
            item = backup_fs.GetByID(remotePath, iterID=backup_fs.fsID(customer_idurl, keyAlias))
            if item:
                versions = item.list_versions(sorted=True, reverse=True)
                if len(versions) > maxBackupsNum:
                    for version in versions[maxBackupsNum:]:
                        item.delete_version(version)
                        backupID = packetid.MakeBackupID(customerGlobalID, remotePath, version, key_alias=keyAlias)
                        backup_rebuilder.RemoveBackupToWork(backupID)
                        backup_fs.DeleteLocalBackup(settings.getLocalBackupsDir(), backupID)
                        backup_matrix.EraseBackupLocalInfo(backupID)
                        backup_matrix.EraseBackupLocalInfo(backupID)
        backup_fs.ScanID(remotePath)
        backup_fs.Calculate(iterID=backup_fs.fsID(customer_idurl, keyAlias))
        SaveFSIndex(customer_idurl, keyAlias)
        # TODO: check used space, if we have over use - stop all tasks immediately
    elif result == 'abort':
        DeleteBackup(backupID)
    if len(tasks()) == 0:
        # do we really need to restart backup_monitor after each backup?
        # if we have a lot tasks started this will produce a lot unneeded actions
        # will be smarter to restart it once we finish all tasks
        # because user will probably leave BitDust working after starting a long running operations
        from bitdust.storage import backup_monitor
        if _Debug:
            lg.out(_DebugLevel, 'backup_control.OnJobDone restarting backup_monitor() machine because no tasks left')
        backup_monitor.A('restart')
    reactor.callLater(0, RunTask)  # @UndefinedVariable
    reactor.callLater(0, FireTaskFinishedCallbacks, remotePath, version, result)  # @UndefinedVariable


def OnJobFailed(backupID, err):
    lg.err('job failed [%s] : %s' % (backupID, err))
    jobs().pop(backupID)


def OnTaskFailed(pathID, result):
    """
    Called when backup process get failed somehow.
    """
    lg.err('task failed [%s] with result "%s", %d more tasks' % (pathID, result, len(tasks())))
    reactor.callLater(0, RunTask)  # @UndefinedVariable
    reactor.callLater(0, FireTaskFinishedCallbacks, pathID, None, result)  # @UndefinedVariable


def OnBackupBlockReport(backupID, blockNum, result):
    """
    Called for every finished block during backup process.

    :param newblock: this is a ``p2p.encrypted_block.encrypted_block`` instance
    :param num_suppliers: number of suppliers which is used for that backup
    """
    backup_matrix.LocalBlockReport(backupID, blockNum, result)


def OnNewDataPrepared():
    data_sender.A('new-data')


def OnTaskExecutedCallback(result):
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.OnTaskExecuted %s : %s' % (result[0], result[1]))
    return result


def OnTaskFailedCallback(result):
    lg.err('pathID: %s, error: %s' % (result[0], result[1]))
    return result


#------------------------------------------------------------------------------


def AddTaskStartedCallback(pathID, callback):
    """
    You can catch a moment when given ``Task`` were started.

    Call this method to provide a callback method to handle.
    """
    global _TaskStartedCallbacks
    pathID = global_id.CanonicalID(pathID)
    if pathID not in _TaskStartedCallbacks:
        _TaskStartedCallbacks[pathID] = []
    _TaskStartedCallbacks[pathID].append(callback)


def AddTaskFinishedCallback(pathID, callback):
    """
    You can also catch a moment when the whole ``Job`` is done and backup
    process were finished or failed.
    """
    global _TaskFinishedCallbacks
    pathID = global_id.CanonicalID(pathID)
    if pathID not in _TaskFinishedCallbacks:
        _TaskFinishedCallbacks[pathID] = []
    _TaskFinishedCallbacks[pathID].append(callback)


def FireTaskStartedCallbacks(pathID, version):
    """
    This runs callbacks for given path ID when that ``Job`` is started.
    """
    global _TaskStartedCallbacks
    for cb in _TaskStartedCallbacks.get(pathID, []):
        cb(pathID, version)
    _TaskStartedCallbacks.pop(pathID, None)


def FireTaskFinishedCallbacks(remotePath, version, result):
    """
    This runs callbacks for given path ID when that ``Job`` is done or failed.
    """
    global _TaskFinishedCallbacks
    for cb in _TaskFinishedCallbacks.get(remotePath, []):
        cb(remotePath, version, result)
    _TaskFinishedCallbacks.pop(remotePath, None)


#------------------------------------------------------------------------------


def StartSingle(pathID, localPath=None, keyID=None):
    """
    A high level method to start a backup of single file or folder.
    """
    from bitdust.storage import backup_monitor
    t = PutTask(pathID=pathID, localPath=localPath, keyID=keyID)
    reactor.callLater(0, RunTask)  # @UndefinedVariable
    reactor.callLater(0, backup_monitor.A, 'restart')  # @UndefinedVariable
    return t


def StartRecursive(pathID, keyID=None):
    """
    A high level method to start recursive backup of given path.

    This is will traverse all paths below this ID in the 'tree' and add
    tasks for them.
    """
    pathID = global_id.CanonicalID(pathID)
    from bitdust.storage import backup_monitor
    startedtasks = []

    def visitor(_pathID, path, info):
        if info.type == backup_fs.FILE:
            if _pathID.startswith(pathID):
                t = PutTask(pathID=pathID, localPath=path, keyID=keyID)
                startedtasks.append(t)

    backup_fs.TraverseByID(visitor)
    reactor.callLater(0, RunTask)  # @UndefinedVariable
    reactor.callLater(0, backup_monitor.A, 'restart')  # @UndefinedVariable
    if _Debug:
        lg.out(_DebugLevel, 'backup_control.StartRecursive %s  :  %d tasks started' % (pathID, len(startedtasks)))
    return startedtasks


#------------------------------------------------------------------------------


def IsBackupInProcess(backupID):
    """
    Return True if given backup ID is running and that "job" exists.
    """
    backupID = global_id.CanonicalID(backupID)
    return backupID in jobs()


def IsPathInProcess(pathID):
    """
    Return True if some backups is running at the moment of given path.
    """
    pathID = global_id.CanonicalID(pathID)
    for backupID in jobs().keys():
        if backupID.startswith(pathID):
            return True
    return False


def FindRunningBackup(pathID=None, customer=None):
    if pathID:
        pathID = global_id.CanonicalID(pathID)
    result = set()
    for backupID in jobs().keys():
        if pathID:
            if backupID.count(pathID):
                result.add(backupID)
                continue
        if customer:
            if backupID.count(customer + ':'):
                result.add(backupID)
                continue
    return list(result)


def HasRunningBackup():
    """
    Return True if at least one backup is running right now.
    """
    return len(jobs()) > 0


def AbortRunningBackup(backupID):
    """
    Call ``abort()`` method of ``p2p.backup.backup`` object - abort the running backup.
    """
    backupID = global_id.CanonicalID(backupID)
    if IsBackupInProcess(backupID):
        jobs()[backupID].abort()
        return True
    return False


def AbortAllRunningBackups():
    """
    Abort all running backups :-).
    """
    for backupObj in jobs().values():
        backupObj.abort()


def ListRunningBackups():
    """
    List backup IDs of currently running jobs.
    """
    return list(jobs().keys())


def GetRunningBackupObject(backupID):
    """
    Return an instance of ``p2p.backup.backup`` class - a running backup object,
    or None if that ID is not exist in the jobs dictionary.
    """
    backupID = global_id.CanonicalID(backupID)
    return jobs().get(backupID, None)


#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------


def test():
    """
    For tests.
    """
    #    backup_fs.Calculate()
    #    print backup_fs.counter()
    #    print backup_fs.numberfiles()
    #    print backup_fs.sizefiles()
    #    print backup_fs.sizebackups()
    pprint.pprint(backup_fs.fsID())
    pprint.pprint(backup_fs.fs())
    print(backup_fs.GetByID('0'))
    # pprint.pprint(backup_fs.WalkByID('0/0/2/19/F20140106100849AM'))
    # for pathID, localPath, item in backup_fs.IterateIDs():
    #     print pathID, misc.unicode_to_str_safe(localPath)
    # backup_fs.TraverseByIDSorted(lambda x, y, z: sys.stdout.write('%s %s\n' % (x,misc.unicode_to_str_safe(y))))


def test2():
    """
    For tests.
    """
    # reactor.callLater(1, StartDirRecursive, 'c:/temp')
    reactor.run()  # @UndefinedVariable


#------------------------------------------------------------------------------

if __name__ == '__main__':
    bpio.init()
    lg.set_debug_level(20)
    settings.init()
    tmpfile.init(settings.getTempDir())
    contactsdb.init()
    eccmap.init()
    key.InitMyKey()
    init()
    test()
    settings.shutdown()



from node import blocknf, digest_block
import threading
from libs import node, logger, client
import time
import dbhandler
import socks
from connections import send, receive
from decimal import Decimal
import mempool as mp
from difficulty import difficulty
from libs import client


def sendsync(sdef, peer_ip, status, node):
    """ Save peer_ip to peerlist and send `sendsync`

    :param sdef: socket object
    :param peer_ip: IP of peer synchronization has been completed with
    :param status: Status synchronization was completed in/as

    Log the synchronization status
    Save peer IP to peers list if applicable
    Wait for database to unlock
    Send `sendsync` command via socket `sdef`

    returns None
    """
    # TODO: ERROR, does **not** save anything. code or comment wrong.
    node.logger.app_log.debug(f'Outbound: Synchronization with {peer_ip} finished after: {status}, sending new sync request')
    time.sleep(int(node.pause))
    while node.db_lock.locked():
        if node.IS_STOPPING:
            return
        time.sleep(int(node.pause))
    send(sdef, 'sendsync')


def worker(host, port, node):
    logger = node.logger

    this_client = f'{host}:{port}'

    if node.IS_STOPPING:
        return

    dict_ip = {'ip': host}
    node.plugin_manager.execute_filter_hook('peer_ip', dict_ip)
    client_instance_worker = client.Client()

    if node.peers.is_banned(host) or dict_ip['ip'] == 'banned':
        node.logger.app_log.warning(f"IP {host} is banned, won't connect")
        return

    timeout_operation = 60  # timeout
    timer_operation = time.time()  # start counting

    # print('started Bismuth peer worker', threading.current_thread(), host, port)

    try:

        s = socks.socksocket()

        if node.tor:
            s.setproxy(socks.PROXY_TYPE_SOCKS5, '127.0.0.1', 9050)
        # s.setblocking(0)
        s.connect((host, port))
        node.logger.app_log.debug(f'Outbound: Connected to {this_client}')
        client_instance_worker.connected = True

        # communication starter

        send(s, 'version')
        send(s, node.version)

        data = receive(s)

        if data == 'ok':
            node.logger.app_log.debug(f'Outbound: Node protocol version of {this_client} matches our client')
        else:
            raise ValueError(f'Outbound: Node protocol version of {this_client} mismatch')

        send(s, 'getversion')
        peer_version = receive(s)
        if peer_version not in node.version_allow:
            raise ValueError(f'Outbound: Incompatible peer version {peer_version} from {this_client}')

        send(s, 'hello')

        # communication starter

    except Exception as e:
        node.logger.app_log.debug(f'Could not connect to {this_client}: {e}')
        return  # can return here, because no lists are affected yet

    node.peers.store_mainnet(host, peer_version)
    try:
        peer_ip = s.getpeername()[0]
    except:
        # Should not happen, extra safety
        node.logger.app_log.warning('Outbound: Transport endpoint was not connected')
        return

    if this_client not in node.peers.connection_pool:
        node.peers.append_client(this_client)
        node.logger.app_log.debug(f'Connected to {this_client}')
        node.logger.app_log.debug(f'Current active pool: {node.peers.connection_pool}')

    # if not node.peers.is_banned(host) and node.peers.version_allowed(host, node.version_allow) and not node.IS_STOPPING:
    #     db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)

    while not node.peers.is_banned(host) and node.peers.version_allowed(host, node.version_allow) and not node.IS_STOPPING:
        try:
            #ensure_good_peer_version(host)

            data = receive(s)  # receive data, one and the only root point
            # print(data)

            if data == 'peers':
                subdata = receive(s)  # dict of "ip":"port"
                node.peers.peersync(subdata)

            elif data == 'sync':
                if not time.time() <= timer_operation + timeout_operation:
                    timer_operation = time.time()  # reset timer

                try:
                    while len(node.syncing) >= 3:
                        if node.IS_STOPPING:
                            return
                        time.sleep(int(node.pause))

                    node.syncing.append(peer_ip)
                    # sync start

                    # send block height, receive block height
                    send(s, 'blockheight')

                    node.logger.app_log.debug(f'Outbound: Sending block height to compare: {node.hdd_block}')
                    # append zeroes to get static length
                    send(s, node.hdd_block)

                    received_block_height = receive(s)  # receive node's block height
                    node.logger.app_log.debug(f'Outbound: Node {peer_ip} is at block height: {received_block_height}')

                    if int(received_block_height) < node.hdd_block:
                        node.logger.app_log.warning(f'Outbound: We have a higher block ({node.hdd_block}) than {peer_ip} ({received_block_height}), sending')

                        data = receive(s)  # receive client's last block_hash

                        # send all our followup hashes
                        node.logger.app_log.debug(f'Outbound: Will seek the following block: {data}')

                        # consensus pool 2 (active connection)
                        consensus_blockheight = int(received_block_height)
                        node.peers.consensus_add(peer_ip, consensus_blockheight, s, node.hdd_block)
                        # consensus pool 2 (active connection)

                        db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)
                        client_block = db_handler_instance.block_height_from_hash(data)
                        db_handler_instance.close()

                        if not client_block:
                            node.logger.app_log.warning(f'Outbound: Block {data[:8]} of {peer_ip} not found')
                            if node.full_ledger:
                                send(s, 'blocknf')
                            else:
                                send(s, 'blocknfhb')
                            send(s, data)

                            if node.peers.warning(s, peer_ip, 'Forked', 1):
                                raise ValueError(f'{peer_ip} is banned')

                        else:
                            node.logger.app_log.warning(f'Outbound: Node is at block {client_block}')  # now check if we have any newer

                            if node.hdd_hash == data or not node.egress:
                                if not node.egress:
                                    node.logger.app_log.warning(f'Outbound: Egress disabled for {peer_ip}')
                                    time.sleep(int(node.pause))  # reduce CPU usage
                                else:
                                    node.logger.app_log.debug(f'Outbound: Node {peer_ip} has the latest block')
                                    # TODO: this is unlikely to happen due to conditions above, consider removing
                                send(s, 'nonewblk')

                            else:
                                db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)
                                blocks_fetched = db_handler_instance.blocksync(client_block)
                                db_handler_instance.close()

                                node.logger.app_log.debug(f'Outbound: Selected {blocks_fetched}')

                                send(s, 'blocksfnd')

                                confirmation = receive(s)

                                if confirmation == 'blockscf':
                                    node.logger.app_log.debug('Outbound: Client confirmed they want to sync from us')
                                    send(s, blocks_fetched)

                                elif confirmation == 'blocksrj':
                                    node.logger.app_log.debug("Outbound: Client rejected to sync from us because we're dont have the latest block")

                    elif int(received_block_height) >= node.hdd_block:
                        if int(received_block_height) == node.hdd_block:
                            node.logger.app_log.debug(f'Outbound: We have the same block as {peer_ip} ({received_block_height}), hash will be verified')
                        else:
                            node.logger.app_log.warning(f'Outbound: We have a lower block ({node.hdd_block}) than {peer_ip} ({received_block_height}), hash will be verified')

                        node.logger.app_log.debug(f'Outbound: block_hash to send: {node.hdd_hash}')
                        send(s, node.hdd_hash)

                        #ensure_good_peer_version(host)

                        # consensus pool 2 (active connection)
                        consensus_blockheight = int(received_block_height)  # str int to remove leading zeros
                        node.peers.consensus_add(peer_ip, consensus_blockheight, s, node.hdd_block)
                        # consensus pool 2 (active connection)

                except Exception as e:
                    node.logger.app_log.warning(f'Outbound: Sync failed {e}')
                finally:
                    node.syncing.remove(peer_ip)

            elif data == 'blocknfhb':  # one of the possible outcomes
                block_hash_delete = receive(s)
                # print peer_ip
                # if max(consensus_blockheight_list) == int(received_block_height):
                if int(received_block_height) == node.peers.consensus_max:

                    db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)
                    blocknf(node, block_hash_delete, peer_ip, db_handler_instance, hyperblocks=True)
                    db_handler_instance.close()

                    if node.peers.warning(s, peer_ip, 'Rollback', 2):
                        raise ValueError(f'{peer_ip} is banned')

                sendsync(s, peer_ip, 'Block not found', node)

            elif data == 'blocknf':  # one of the possible outcomes
                block_hash_delete = receive(s)
                # print peer_ip
                # if max(consensus_blockheight_list) == int(received_block_height):
                if int(received_block_height) == node.peers.consensus_max:

                    db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)
                    blocknf(node, block_hash_delete, peer_ip, db_handler_instance)
                    db_handler_instance.close()

                    if node.peers.warning(s, peer_ip, 'Rollback', 2):
                        raise ValueError(f'{peer_ip} is banned')

                sendsync(s, peer_ip, 'Block not found', node)

            elif data == 'blocksfnd':
                node.logger.app_log.debug(f'Outbound: Node {peer_ip} has the block(s)')  # node should start sending txs in this step

                # node.logger.app_log.debug("Inbound: Combined segments: " + segments)
                # print peer_ip
                if node.db_lock.locked():
                    node.logger.app_log.warning(f'Skipping sync from {peer_ip}, syncing already in progress')

                else:
                    if int(node.last_block_timestamp) < (time.time() - 600):
                        block_req = node.peers.consensus_most_common
                        node.logger.app_log.warning('Most common block rule triggered')

                    else:
                        block_req = node.peers.consensus_max
                        node.logger.app_log.warning('Longest chain rule triggered')

                    #ensure_good_peer_version(host)

                    if int(received_block_height) >= block_req and int(received_block_height) > node.last_block:
                        try:  # they claim to have the longest chain, things must go smooth or ban
                            node.logger.app_log.warning(f'Confirming to sync from {peer_ip}')

                            send(s, 'blockscf')
                            segments = receive(s)
                            #ensure_good_peer_version(host)

                        except:
                            if node.peers.warning(s, peer_ip, 'Failed to deliver the longest chain', 2):
                                raise ValueError(f'{peer_ip} is banned')
                        else:
                            db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)
                            digest_block(node, segments, s, peer_ip, db_handler_instance)
                            db_handler_instance.close()

                            # receive theirs
                    else:
                        send(s, 'blocksrj')
                        node.logger.app_log.warning(f'Inbound: Distant peer {peer_ip} is at {received_block_height}, should be at least {max(block_req,node.last_block+1)}')

                sendsync(s, peer_ip, 'Block found', node)

                # block_hash validation end

            elif data == 'nonewblk':
                # send and receive mempool
                # print('nonewblk', mp.MEMPOOL, id(mp.MEMPOOL), getattr(mp.MEMPOOL, 'sendable', '?'), threading.current_thread())
                if mp.MEMPOOL.sendable(peer_ip):
                    mempool_txs = mp.MEMPOOL.tx_to_send(peer_ip)
                    # node.logger.app_log.debug("Outbound: Extracted from the mempool: " + str(mempool_txs))  # improve: sync based on signatures only
                    # if len(mempool_txs) > 0: #wont sync mempool until we send something, which is bad
                    # send own
                    send(s, 'mempool')
                    send(s, mempool_txs)
                    # send own
                    # receive theirs
                    segments = receive(s)

                    db_handler_instance = dbhandler.DbHandler(node.index_db, node.ledger_path, node.hyper_path, node.ram, node.ledger_ram_file, logger)
                    node.logger.app_log.debug(mp.MEMPOOL.merge(segments, peer_ip, db_handler_instance.c, True))
                    db_handler_instance.close()

                    # receive theirs
                    # Tell the mempool we just send our pool to a peer
                    mp.MEMPOOL.sent(peer_ip)
                sendsync(s, peer_ip, 'No new block', node)

            elif data == 'hyperlane':
                pass

            else:
                if data == '*':
                    raise ValueError('Broken pipe')
                raise ValueError(f'Unexpected error, received: {str(data)[:32]}')

        except Exception as e:
            """
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            """
            import traceback
            traceback.print_exc()

            # remove from active pool
            node.peers.remove_client(this_client)
            node.logger.app_log.warning(f'Outbound: Disconnected from {this_client}: {e}')
            # remove from active pool

            # remove from consensus 2
            node.peers.consensus_remove(peer_ip)
            # remove from consensus 2

            node.logger.app_log.debug(f'Connection to {this_client} terminated due to {e}')
            node.logger.app_log.debug(f'---thread {threading.currentThread()} ended---')

            # properly end the connection
            s.close()

            # properly end the connection
            if node.debug:
                raise  # major debug client
            else:
                node.logger.app_log.warning(f'Ending thread, because {e}')
                return

    if not node.peers.version_allowed(host, node.version_allow):
        node.logger.app_log.warning(f'Outbound: Ending thread, because {host} has too old a version: {node.peers.ip_to_mainnet[host]}')



#!/usr/bin/env python
# broadcaster_node.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (broadcaster_node.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: broadcaster_node.

.. role:: red

BitDust broadcaster_node() Automat

EVENTS:
    * :red:`broadcast-message-received`
    * :red:`broadcaster-connected`
    * :red:`broadcaster-disconnected`
    * :red:`init`
    * :red:`lookup-failed`
    * :red:`new-broadcaster-connected`
    * :red:`new-outbound-message`
    * :red:`reconnect`
    * :red:`shutdown`
    * :red:`timer-1min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import time
import json

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.main import config

from bitdust.p2p import online_status
from bitdust.p2p import commands

from bitdust.transport import callback

from bitdust.p2p import p2p_service

#------------------------------------------------------------------------------

_BroadcasterNode = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _BroadcasterNode
    if event is None and not args:
        return _BroadcasterNode
    if _BroadcasterNode is None:
        # set automat name and starting state here
        _BroadcasterNode = BroadcasterNode('broadcaster_node', 'AT_STARTUP', _DebugLevel, _Debug)
    if event is not None:
        _BroadcasterNode.automat(event, *args, **kwargs)
    return _BroadcasterNode


#------------------------------------------------------------------------------


class BroadcasterNode(automat.Automat):
    """
    This class implements all the functionality of the ``broadcaster_node()``
    state machine.
    """

    timers = {
        'timer-1min': (60, ['BROADCASTING']),
    }

    def init(self):
        self.max_broadcasters = config.conf().getInt('services/broadcasting/max-broadcast-connections', 3)
        self.connected_broadcasters = []
        self.messages_sent = {}
        # self.messages_acked = {}
        self.last_success_action_time = None
        self.listeners = {}
        self.incoming_broadcast_message_callback = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'BROADCASTERS?'
                self.doInit(*args, **kwargs)
                self.doStartBroadcastersLookup(*args, **kwargs)
        #---BROADCASTERS?---
        elif self.state == 'BROADCASTERS?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doEraseBroadcasters(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'broadcaster-connected' and not self.isMoreNeeded(*args, **kwargs):
                self.state = 'BROADCASTING'
                self.doAddBroadcaster(*args, **kwargs)
            elif event == 'lookup-failed' and not self.isAnyBroadcasters(*args, **kwargs):
                self.state = 'OFFLINE'
            elif event == 'broadcaster-connected' and self.isMoreNeeded(*args, **kwargs):
                self.doAddBroadcaster(*args, **kwargs)
                self.doStartBroadcastersLookup(*args, **kwargs)
            elif event == 'lookup-failed' and self.isAnyBroadcasters(*args, **kwargs):
                self.state = 'BROADCASTING'
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'reconnect' or event == 'new-broadcaster-connected':
                self.state = 'BROADCASTERS?'
                self.doAddBroadcaster(*args, **kwargs)
                self.doStartBroadcastersLookup(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---BROADCASTING---
        elif self.state == 'BROADCASTING':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doEraseBroadcasters(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'new-broadcaster-connected':
                self.doAddBroadcaster(*args, **kwargs)
            elif event == 'broadcast-message-received':
                self.doCheckAndSendForward(*args, **kwargs)
            elif event == 'broadcaster-disconnected':
                self.doRemoveBroadcaster(*args, **kwargs)
            elif event == 'new-outbound-message':
                self.doBroadcastMessage(*args, **kwargs)
            elif event == 'timer-1min' and self.isLineActive(*args, **kwargs):
                self.doTestReconnectBroadcasters(*args, **kwargs)
            elif event == 'timer-1min' and not self.isLineActive(*args, **kwargs):
                self.state = 'OFFLINE'
                self.doEraseBroadcasters(*args, **kwargs)
        return None

    def isLineActive(self, *args, **kwargs):
        """
        Condition method.
        """
        for idurl in self.connected_broadcasters:
            if online_status.isOnline(idurl):
                # if at least one broadcaster is connected
                # we assume the line is still active
                return True
        if not self.last_success_action_time:
            return False
        return time.time() - self.last_success_action_time > 5*60

    def isAnyBroadcasters(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.connected_broadcasters) > 0

    def isMoreNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.connected_broadcasters) < self.max_broadcasters

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.incoming_broadcast_message_callback = args[0]
        callback.append_inbox_callback(self._on_inbox_packet)

    def doStartBroadcastersLookup(self, *args, **kwargs):
        """
        Action method.
        """
        from broadcast import broadcasters_finder
        broadcasters_finder.A('start', (
            self.automat,
            {
                'action': 'route',
            },
            list(self.connected_broadcasters),
        ))

    def doAddBroadcaster(self, *args, **kwargs):
        """
        Action method.
        """
        if not args:
            return
        if args[0] in self.connected_broadcasters:
            lg.out(_DebugLevel, 'broadcaster_node.doAddBroadcaster SKIP, %s already connected as broadcaster' % args[0])
            return
        if args[0] in self.listeners:
            lg.out(_DebugLevel, 'broadcaster_node.doAddBroadcaster SKIP, %s already connected as listener' % args[0])
        self.connected_broadcasters.append(args[0])
        self.last_success_action_time = time.time()
        if _Debug:
            lg.out(_DebugLevel, 'broadcaster_node.doAddBroadcaster %s joined, %d total connected' % (args[0], len(self.connected_broadcasters)))

    def doRemoveBroadcaster(self, *args, **kwargs):
        """
        Action method.
        """
        if args[0] not in self.connected_broadcasters:
            lg.warn('%s is not connected' % args[0])
            return
        self.connected_broadcasters.remove(args[0])
        if _Debug:
            lg.out(_DebugLevel, 'broadcaster_node.doRemoveBroadcaster %s now disconnected' % args[0])

    def doEraseBroadcasters(self, *args, **kwargs):
        """
        Action method.
        """
        self.connected_broadcasters = []

    def doCheckAndSendForward(self, *args, **kwargs):
        """
        Action method.
        """
        from broadcast import broadcast_service
        msg, newpacket = args[0]
        msgid = msg['id']
        if _Debug:
            lg.out(_DebugLevel, 'broadcaster_node.doCheckAndSendForward %s' % msgid)
        self.last_success_action_time = time.time()
        # skip broadcasting if this message was already sent
        if msgid in self.messages_sent:
            if _Debug:
                lg.out(_DebugLevel, '        resent skipped, %s was already sent to my broadcasters' % msgid)


#             if msgid not in self.messages_acked:
#                 p2p_service.SendAck(newpacket, '0')
#             else:
#                 p2p_service.SendAck(newpacket, str(self.messages_acked[msgid]))
            return
        # if some listeners connected - send to them
        for listener_idurl, scope in self.listeners.items():
            lg.out(4, '           test %s:%s for %s' % (listener_idurl, scope, msg['owner']))
            # but check if they really need that message
            # listener can set a scope, so he will get this broadcasting
            # only if creator of that message is listed in scope
            if not scope or msg['owner'] in scope:
                outpacket = broadcast_service.packet_for_listener(listener_idurl, msg)
                p2p_service.SendBroadcastMessage(outpacket)
        # fire broadcast listening callback
        if self.incoming_broadcast_message_callback is not None:
            self.incoming_broadcast_message_callback(msg)
        # finally broadcast further
        for broadcaster_idurl in self.connected_broadcasters:
            outpacket = broadcast_service.packet_for_broadcaster(broadcaster_idurl, msg)
            p2p_service.SendBroadcastMessage(outpacket)
        self.messages_sent[msgid] = int(time.time())

    def doBroadcastMessage(self, *args, **kwargs):
        """
        Action method.
        """
        from broadcast import broadcast_service
        msg, newpacket = args[0]
        msgid = msg['id']
        if _Debug:
            lg.out(_DebugLevel, 'broadcaster_node.doBroadcastMessage %s' % msgid)
        if msgid in self.messages_sent:
            lg.warn('CRITICAL, found same message already broadcasted !!!')
            return
        # if some listeners connected - send to them
        for listener_idurl, scope in self.listeners.items():
            if listener_idurl == newpacket.OwnerID:
                # skip this listener
                continue
            lg.out(4, '           test %s:%s for %s' % (listener_idurl, scope, msg['owner']))
            # but check if they really need that message
            # listener can set a scope, so he will get this broadcasting
            # only if creator of that message is listed in scope
            if not scope or msg['owner'] in scope:
                outpacket = broadcast_service.packet_for_listener(listener_idurl, msg)
                p2p_service.SendBroadcastMessage(outpacket)
        # fire broadcast listening callback
        if self.incoming_broadcast_message_callback is not None:
            self.incoming_broadcast_message_callback(msg)
        for broadcaster_idurl in self.connected_broadcasters:
            outpacket = broadcast_service.packet_for_broadcaster(broadcaster_idurl, msg)
            p2p_service.SendBroadcastMessage(outpacket)
        self.messages_sent[msgid] = int(time.time())

    def doTestReconnectBroadcasters(self, *args, **kwargs):
        """
        Action method.
        """

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.incoming_broadcast_message_callback = None
        callback.remove_inbox_callback(self._on_inbox_packet)
        self.destroy()
        global _BroadcasterNode
        del _BroadcasterNode
        _BroadcasterNode = None

    #------------------------------------------------------------------------------

    def add_listener(self, listener_idurl, scope_as_string):
        try:
            scope = json.loads(scope_as_string)
        except:
            lg.exc(scope_as_string)
            return
        if listener_idurl in self.connected_broadcasters:
            lg.warn('%s already connected as broadcaster, can not set it as listener' % listener_idurl)
            return
        self.listeners[listener_idurl] = scope

    def remove_listener(self, listener_idurl):
        if listener_idurl not in self.listeners:
            lg.warn('%s not in listeners at the moment')
            return
        self.listeners.pop(listener_idurl)

    #------------------------------------------------------------------------------

    def _on_inbox_packet(self, newpacket, info, status, error_message):
        if newpacket.Command == commands.Broadcast():
            from broadcast import broadcast_service
            msg = broadcast_service.read_message_from_packet(newpacket)
            if not msg:
                return False
            if msg['id'] in self.messages_sent:
                if _Debug:
                    lg.out(_DebugLevel, 'broadcaster_node._on_inbox_packet SKIPPED, %s already broadcasted' % msg['id'])
                return True
            if msg['owner'] in self.listeners:
                if msg['owner'] in self.connected_broadcasters:
                    lg.warn('%s present in both lists: listeners and broadcasters!!!' % msg['owner'])
                    return True
                # message from listener - start broadcasting
                self.automat('new-outbound-message', (msg, newpacket))
            else:
                self.automat('broadcast-message-received', (msg, newpacket))
            return True
        return False



#!/usr/bin/python
# message_keeper.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (message_keeper.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: message_keeper

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.interface import api_web_socket

from bitdust.crypt import my_keys

from bitdust.contacts import identitycache

from bitdust.stream import message

from bitdust.chat import message_database

from bitdust.userid import global_id

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'message_keeper.init')
    message.consume_messages(
        consumer_callback_id='message_keeper',
        callback=on_consume_user_messages,
        direction=None,
        message_types=[
            'private_message',
            'group_message',
        ],
        reset_callback=False,
    )


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'message_keeper.shutdown')
    message.clear_consumer_callbacks(consumer_callback_id='message_keeper')


#------------------------------------------------------------------------------


def on_consume_user_messages(json_messages):
    for json_message in json_messages:
        try:
            msg_type = json_message.get('type', '')
            packet_id = json_message['packet_id']
            sender_id = json_message['from']
            recipient_id = json_message['to']
            direction = json_message['dir']
            msg_data = json_message['data']
        except:
            lg.exc()
            continue
        cache_message(
            data=msg_data,
            message_id=packet_id,
            sender_id=sender_id,
            recipient_id=recipient_id,
            message_type=msg_type,
            direction=direction,
        )
    return False


#------------------------------------------------------------------------------


def cache_message(data, message_id, sender_id, recipient_id, message_type=None, direction=None):
    if _Debug:
        lg.args(_DebugLevel, message_id=message_id, sender_id=sender_id, recipient_id=recipient_id, message_type=message_type)
    if message_type == 'private_message':
        if not my_keys.is_key_registered(sender_id):
            sender_idurl = global_id.glob2idurl(sender_id)
            known_ident = identitycache.FromCache(sender_idurl)
            if not known_ident:
                lg.warn('sender identity %r was not cached, not possible to store message locally' % sender_idurl)
                return False
            if not my_keys.register_key(sender_id, known_ident.getPublicKey()):
                lg.err('failed to register known public key of the sender: %r' % sender_id)
                return False
        if not my_keys.is_key_registered(recipient_id):
            recipient_idurl = global_id.glob2idurl(recipient_id)
            known_ident = identitycache.FromCache(recipient_idurl)
            if not known_ident:
                lg.warn('recipient identity %r was not cached, not possible to store message locally' % recipient_idurl)
                return False
            if not my_keys.register_key(recipient_id, known_ident.getPublicKey()):
                lg.err('failed to register known public key of the recipient: %r' % recipient_id)
                return False
        return store_message(data, message_id, sender_id, recipient_id, message_type, direction)

    if message_type == 'group_message' or message_type == 'personal_message':
        if not my_keys.is_key_registered(recipient_id):
            lg.err('failed to cache %r because recipient key %r was not registered' % (message_type, recipient_id))
            return False
        return store_message(data, message_id, sender_id, recipient_id, message_type, direction)

    raise Exception('unexpected message type: %r' % message_type)


#------------------------------------------------------------------------------


def store_message(data, message_id, sender_id, recipient_id, message_type=None, direction=None):
    message_json = message_database.insert_message(
        data=data,
        message_id=message_id,
        sender=sender_id,
        recipient=recipient_id,
        message_type=message_type,
        direction=direction,
    )
    if not message_json:
        lg.warn('message %r was not stored' % message_id)
        return message_json
    api_web_socket.on_stream_message(message_json)
    if _Debug:
        lg.out(_DebugLevel, 'message_keeper.store_message [%s]:%s from %r to %r' % (message_type, message_id, sender_id, recipient_id))
    return message_json



#!/usr/bin/python
# raidutils.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (raidutils.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: raidutils.

"""

import array


def build_parity(sds, iters, datasegments, myeccmap, paritysegments, threshold_control=None):
    psds_list = {seg_num: array.array('i') for seg_num in range(myeccmap.paritysegments)}

    for i in range(iters):
        parities = {seg_num: 0 for seg_num in range(myeccmap.paritysegments)}
        for DSegNum in range(datasegments):
            b = next(sds[DSegNum])

            Map = myeccmap.DataToParity[DSegNum]
            for PSegNum in Map:
                if PSegNum > paritysegments:
                    myeccmap.check()
                    raise Exception('eccmap error')

                parities[PSegNum] = parities[PSegNum] ^ b

                if threshold_control:
                    if not threshold_control(1):
                        raise Exception('task cancelled')

        for PSegNum in range(myeccmap.paritysegments):
            psds_list[PSegNum].append(parities[PSegNum])

    for PSegNum in psds_list:
        psds_list[PSegNum].byteswap()

    return psds_list


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]



#!/usr/bin/python
# service_nodes_lookup.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_nodes_lookup.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_nodes_lookup
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return NodesLookupService()


class NodesLookupService(LocalService):

    service_name = 'service_nodes_lookup'
    config_path = 'services/nodes-lookup/enabled'

    def dependent_on(self):
        # TODO:
        # in the future we can use other methods to discover nodes
        # it can be hard-coded list of nodes, a blockchain, some broadcasting solution, etc...
        # then we redefine that logic in lookup_method and observe_method
        return [
            'service_entangled_dht',
            'service_p2p_hookups',
        ]

    def start(self):
        from bitdust.p2p import lookup
        lookup.init(
            lookup_method=lookup.lookup_in_dht,
            observe_method=lookup.observe_dht_node,
            process_method=lookup.process_idurl,
        )
        # lookup.start(count=1, consume=False)
        return True

    def stop(self):
        from bitdust.p2p import lookup
        lookup.shutdown()
        return True



#!/usr/bin/env python
# nickname_holder.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (nickname_holder.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: nickname_holder

.. role:: red


BitDust nickname_holder() Automat

.. raw:: html

    <a href="nickname_holder.png" target="_blank">
    <img src="nickname_holder.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`dht-erase-failed`
    * :red:`dht-erase-success`
    * :red:`dht-read-failed`
    * :red:`dht-read-success`
    * :red:`dht-write-failed`
    * :red:`dht-write-success`
    * :red:`set`
    * :red:`timer-5min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.main import settings

from bitdust.userid import my_id
from bitdust.userid import id_url

from bitdust.dht import dht_service
from bitdust.dht import dht_records

#------------------------------------------------------------------------------

_NicknameHolder = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _NicknameHolder
    if event is None:
        return _NicknameHolder
    if _NicknameHolder is None:
        # set automat name and starting state here
        _NicknameHolder = NicknameHolder(
            name='nickname_holder',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _NicknameHolder.automat(event, *args, **kwargs)
    return _NicknameHolder


def Destroy():
    """
    Destroy nickname_holder() automat and remove its instance from memory.
    """
    global _NicknameHolder
    if _NicknameHolder is None:
        return
    _NicknameHolder.destroy()
    del _NicknameHolder
    _NicknameHolder = None


#------------------------------------------------------------------------------


class NicknameHolder(automat.Automat):
    """
    This class implements all the functionality of the ``nickname_holder()``
    state machine.
    """

    timers = {
        'timer-5min': (300, ['READY']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.nickname = None
        self.key = None
        self.dht_read_defer = None
        self.result_callbacks = []

    def add_result_callback(self, cb):
        self.result_callbacks.append(cb)

    def remove_result_callback(self, cb):
        self.result_callbacks.remove(cb)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'set':
                self.state = 'DHT_READ'
                self.Attempts = 0
                self.doSetNickname(*args, **kwargs)
                self.doMakeKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
        #---READY---
        elif self.state == 'READY':
            if event == 'timer-5min':
                self.state = 'DHT_READ'
                self.doSetNickname(*args, **kwargs)
                self.doMakeKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'set':
                self.state = 'DHT_ERASE'
                self.doDHTEraseKey(*args, **kwargs)
                self.doSetNickname(*args, **kwargs)
        #---DHT_READ---
        elif self.state == 'DHT_READ':
            if event == 'dht-read-success' and self.isMyOwnKey(*args, **kwargs):
                self.state = 'READY'
                self.doReportNicknameOwn(*args, **kwargs)
            elif event == 'dht-read-failed':
                self.state = 'DHT_WRITE'
                self.doDHTWriteKey(*args, **kwargs)
            elif event == 'dht-read-success' and not self.isMyOwnKey(*args, **kwargs):
                self.doReportNicknameExist(*args, **kwargs)
                self.doNextKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'set':
                self.doSetNickname(*args, **kwargs)
                self.doMakeKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
        #---DHT_WRITE---
        elif self.state == 'DHT_WRITE':
            if event == 'dht-write-failed' and self.Attempts > 5:
                self.state = 'READY'
                self.Attempts = 0
                self.doReportNicknameFailed(*args, **kwargs)
            elif event == 'dht-write-failed' and self.Attempts <= 5:
                self.state = 'DHT_READ'
                self.Attempts += 1
                self.doNextKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'dht-write-success':
                self.state = 'READY'
                self.Attempts = 0
                self.doReportNicknameRegistered(*args, **kwargs)
            elif event == 'set':
                self.state = 'DHT_READ'
                self.doSetNickname(*args, **kwargs)
                self.doMakeKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
        #---DHT_ERASE---
        elif self.state == 'DHT_ERASE':
            if event == 'dht-erase-success' or event == 'dht-erase-failed':
                self.state = 'DHT_READ'
                self.doMakeKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'set':
                self.state = 'DHT_READ'
                self.doSetNickname(*args, **kwargs)
                self.doMakeKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
        return None

    def isMyOwnKey(self, *args, **kwargs):
        """
        Condition method.
        """
        return id_url.to_bin(args[0]) == my_id.getIDURL().to_bin()

    def doSetNickname(self, *args, **kwargs):
        """
        Action method.
        """
        self.nickname = my_id.getLocalIdentity().getIDName()

    def doMakeKey(self, *args, **kwargs):
        """
        Action method.
        """
        self.key = dht_service.make_key(
            key=self.nickname,
            index=0,
            prefix='nickname',
        )

    def doNextKey(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            key_info = dht_service.split_key(self.key)
            index = int(key_info['index'])
        except:
            lg.exc()
            index = 0
        index += 1
        self.key = dht_service.make_key(
            key=self.nickname,
            index=index,
            prefix='nickname',
        )

    def doDHTReadKey(self, *args, **kwargs):
        """
        Action method.
        """
        if self.dht_read_defer is not None:
            self.dht_read_defer.pause()
            self.dht_read_defer.cancel()
            self.dht_read_defer = None
        d = dht_records.get_nickname(self.key, use_cache=False)
        d.addCallback(self._dht_read_result, self.key)
        d.addErrback(self._dht_read_failed)
        self.dht_read_defer = d

    def doDHTWriteKey(self, *args, **kwargs):
        """
        Action method.
        """
        d = dht_records.set_nickname(self.key, my_id.getIDURL())
        d.addCallback(self._dht_write_result)
        d.addErrback(lambda err: self.automat('dht-write-failed', err))

    def doDHTEraseKey(self, *args, **kwargs):
        """
        Action method.
        """
        d = dht_service.delete_key(self.key)
        d.addCallback(self._dht_erase_result)
        d.addErrback(lambda x: self.automat('dht-erase-failed'))

    def doReportNicknameOwn(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_holder.doReportNicknameOwn : %s with %s' % (self.key, args[0]))
        for cb in self.result_callbacks:
            cb('my own', self.key)

    def doReportNicknameRegistered(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_holder.doReportNicknameRegistered : %s' % self.key)
        for cb in self.result_callbacks:
            cb('registered', self.key)

    def doReportNicknameExist(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_holder.doReportNicknameExist : %s' % self.key)
        for cb in self.result_callbacks:
            cb('exist', self.key)

    def doReportNicknameFailed(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_holder.doReportNicknameFailed : %s with %s' % (self.key, args[0] if args else None))
        for cb in self.result_callbacks:
            cb('failed', self.key)

    def _dht_read_result(self, value, key):
        self.dht_read_defer = None
        if not value:
            self.automat('dht-read-failed')
            return
        try:
            v = id_url.field(value['idurl'])
        except:
            if _Debug:
                lg.out(_DebugLevel, '%r' % value)
            lg.exc()
            self.automat('dht-read-failed')
            return
        self.automat('dht-read-success', v)

    def _dht_read_failed(self, x):
        self.dht_read_defer = None
        self.automat('dht-read-failed', x)

    def _dht_write_result(self, nodes):
        if len(nodes) > 0:
            self.automat('dht-write-success')
        else:
            self.automat('dht-write-failed', nodes)

    def _dht_erase_result(self, result):
        if result is None:
            self.automat('dht-erase-failed')
        else:
            self.automat('dht-erase-success')


#------------------------------------------------------------------------------


def main():
    from twisted.internet import reactor  # @UnresolvedImport
    lg.set_debug_level(24)
    settings.init()
    my_id.init()
    dht_service.init(settings.getDHTPort())
    reactor.callWhenRunning(A, 'init', sys.argv[1])  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# bpio.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (bpio.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: bpio.

This module is for simple BitDust routines that do not require importing any of our code.:
    - print logs
    - file system IO operations
    - pack/unpack lists and dictionaries into strings
    - some methods to operate with file system paths
    - list Linux mount points and Windows drives
    - methods to manage system processes

Most used method here is ``log`` - prints a log string.

TODO: need to do some refactoring here
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open

#------------------------------------------------------------------------------

import os
import sys
import imp
import platform
import glob
import re
import shutil

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

from bitdust.system import local_fs

#------------------------------------------------------------------------------

LocaleInstalled = False
PlatformInfo = None
X11isRunning = None

#------------------------------------------------------------------------------


def init():
    """
    This method must be called firstly, before any logs will be printed.

    This installs a system locale, so all output messages will have a
    correct encoding.
    """
    InstallLocale()
    if Linux() or Mac():
        lg.setup_unbuffered_stdout()
        lg.setup_unbuffered_stderr()
    if Windows():
        import encodings.idna  # nopycln: import  # @UnusedImport


def shutdown():
    """
    This is the last method to be invoked by the program before main process
    will stop.
    """
    lg.restore_original_stdout()
    lg.restore_original_stderr()
    lg.close_log_file()
    lg.close_intercepted_log_file()
    lg.disable_logs()


def InstallLocale():
    """
    Here is a small trick to refresh current default encoding.
    """
    global LocaleInstalled
    if LocaleInstalled:
        return False
    try:
        import sys
        if sys.version_info[0] == 2:
            reload(sys)  # @UndefinedVariable
            if Windows():
                if hasattr(sys, 'setdefaultencoding'):
                    import locale
                    denc = locale.getpreferredencoding()
                    if not denc:
                        sys.setdefaultencoding('UTF8')
                    else:
                        sys.setdefaultencoding(denc)
            else:
                sys.setdefaultencoding('UTF8')
        LocaleInstalled = True
    except:
        lg.exc()
    return LocaleInstalled


def ostype():
    """
    Return current platform: "Linux", "Windows", "Darwin".
    """
    global PlatformInfo
    if PlatformInfo is None:
        PlatformInfo = list(platform.uname())
        if sys.executable == 'android_python' or ('ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ):
            PlatformInfo[0] = 'Android'
    return PlatformInfo[0]


def osversion():
    """
    Return something like: "2.6.32.9-rscloud" or "XP".
    """
    global PlatformInfo
    if PlatformInfo is None:
        PlatformInfo = list(platform.uname())
    return PlatformInfo[2]


def osinfo():
    """
    Return full OS info, like: "Linux-2.6.32.9-rscloud-x86_64-with-
    Ubuntu-12.04-precise" or "Windows-XP-5.1.2600-SP3".
    """
    if Android():
        try:
            u = platform.uname()
            return '%s %s %s' % (strng.to_text(u[0]), strng.to_text(u[2]), strng.to_text(u[3]))
        except:
            return 'Android'
    return strng.to_text(platform.platform()).strip()


def osinfofull():
    """
    Return detailed system info.
    """
    import pprint
    o = ''
    o += '=====================================================\n'
    o += '=====================================================\n'
    o += '=====================================================\n'
    o += 'platform.uname(): ' + strng.to_text(platform.uname()) + '\n'
    try:
        o += '__file__: ' + strng.to_text(__file__) + '\n'
    except:
        o += 'variable __file__ is not defined\n'
    o += 'sys.executable: ' + sys.executable + '\n'
    o += 'os.path.abspath("."): ' + os.path.abspath('.') + '\n'
    o += 'os.path.abspath(sys.argv[0]): ' + os.path.abspath(sys.argv[0]) + '\n'
    o += 'os.path.expanduser("~"): ' + os.path.expanduser('~') + '\n'
    o += 'sys.argv: ' + pprint.pformat(sys.argv) + '\n'
    o += 'sys.path:\n' + pprint.pformat(sys.path) + '\n'
    o += 'os.environ:\n' + pprint.pformat(list(os.environ.items())) + '\n'
    o += '=====================================================\n'
    o += '=====================================================\n'
    o += '=====================================================\n'
    return o


def windows_version():
    """
    Useful to detect current Windows version: XP, Vista, 7 or 8.
    """
    if getattr(sys, 'getwindowsversion', None) is not None:
        return sys.getwindowsversion()[0]  # @UndefinedVariable
    return 0


def Linux():
    """
    Return True if current platform is Linux.
    """
    return ostype() == 'Linux'


def Windows():
    """
    Return True if current platform is Windows.
    """
    return ostype() == 'Windows'


def Mac():
    """
    Return True if current platform is Mac.
    """
    return ostype() == 'Darwin'


def Android():
    """
    Return True if running on Android inside Kivy.
    """
    return ostype() == 'Android'


def isFrozen():
    """
    Return True if BitDust is running from exe, not from sources.
    """
    return main_is_frozen()


def isConsoled():
    """
    Return True if output can be sent to console.
    """
    if getExecutableFilename().count('pythonw.exe'):
        return False
    if getExecutableFilename().count('bitdust-node.exe'):
        return False
    if getExecutableFilename().count('BitDust-node'):
        return False
    if not sys.stdout:
        return False
    return True


#-------------------------------------------------------------------------------


def list_dir_safe(dirpath):
    """
    A safe wrapper around built-in ``os.listdir()`` method.
    """
    try:
        return os.listdir(dirpath)
    except:
        return []


def list_dir_recursive(dirpath):
    """
    Recursively scan files and folders under ``dirpath`` and return them in the
    list.
    """
    r = []
    for name in os.listdir(dirpath):
        full_name = os.path.join(dirpath, name)
        if os.path.isdir(full_name):
            r.extend(list_dir_recursive(full_name))
        else:
            r.append(full_name)
    return r


def traverse_dir_recursive(callback, basepath, relative_path=''):
    """
    Call ``callback`` method for every file and folder under ``basepath``.

    If method ``callback`` can returns False traverse process will not
    go deeper. Useful to count size of the whole folder.
    """
    if os.path.isdir(basepath):
        for name in os.listdir(basepath):
            real_path = os.path.join(basepath, name)
            sub_path = name if relative_path == '' else relative_path + '/' + name
            go_down = callback(real_path, sub_path, name)
            if os.path.isdir(real_path) and go_down:
                traverse_dir_recursive(callback, real_path, sub_path)


def rmdir_recursive(dirpath, ignore_errors=False, pre_callback=None):
    """
    Remove a directory, and all its contents if it is not already empty.

    http://mail.python.org/pipermail/python-
    list/2000-December/060960.html If ``ignore_errors`` is True process
    will continue even if some errors happens. Method ``pre_callback``
    can be used to decide before remove the file.
    """
    counter = 0
    for name in os.listdir(dirpath):
        full_name = os.path.join(dirpath, name)
        # on Windows, if we don't have write permission we can't remove
        # the file/directory either, so turn that on
        if not os.access(full_name, os.W_OK):
            try:
                os.chmod(full_name, 0o600)
            except:
                continue
        if os.path.isdir(full_name):
            counter += rmdir_recursive(full_name, ignore_errors, pre_callback)
        else:
            if pre_callback:
                if not pre_callback(full_name):
                    continue
            if os.path.isfile(full_name):
                if not ignore_errors:
                    os.remove(full_name)
                    counter += 1
                else:
                    try:
                        os.remove(full_name)
                        counter += 1
                    except Exception as exc:
                        lg.err('can not remove file %r : %r' % (full_name, exc))
                        continue
    if pre_callback:
        if not pre_callback(dirpath):
            return counter
    if not ignore_errors:
        os.rmdir(dirpath)
    else:
        try:
            os.rmdir(dirpath)
        except Exception as exc:
            lg.err('can not remove dir %r : %r' % (dirpath, exc))
    return counter


def move_dir_recursive(src, dest, ignore=None):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        if ignore is not None:
            ignored = ignore(src, files)
        else:
            ignored = set()
        for f in files:
            if f not in ignored:
                move_dir_recursive(os.path.join(src, f), os.path.join(dest, f), ignore=ignore)
    else:
        if os.path.isfile(src):
            shutil.copyfile(src, dest)


def getDirectorySize(directory, include_subfolders=True):
    """
    Platform dependent way to calculate folder size.
    """
    if Windows():
        import win32file  # @UnresolvedImport
        import win32con  # @UnresolvedImport
        import pywintypes  # @UnresolvedImport
        DIR_EXCLUDES = set(['.', '..'])
        MASK = win32con.FILE_ATTRIBUTE_DIRECTORY | win32con.FILE_ATTRIBUTE_SYSTEM
        REQUIRED = win32con.FILE_ATTRIBUTE_DIRECTORY
        FindFilesW = win32file.FindFilesW

        def _get_dir_size(path):
            total_size = 0
            try:
                items = FindFilesW(path + r'\*')
            except pywintypes.error as ex:
                return total_size
            for item in items:
                total_size += item[5]
                if item[0] & MASK == REQUIRED and include_subfolders:
                    name = item[8]
                    if name not in DIR_EXCLUDES:
                        total_size += _get_dir_size(path + '\\' + name)
            return total_size

        return _get_dir_size(directory)
    dir_size = 0
    if not include_subfolders:
        for filename in os.listdir(directory):
            filepath = os.path.abspath(os.path.join(directory, filename))
            if os.path.isfile(filepath):
                try:
                    dir_size += os.path.getsize(filepath)
                except:
                    pass
    else:
        for (path, dirs, files) in os.walk(directory):
            for file in files:
                filename = os.path.join(path, file)
                if os.path.isfile(filename):
                    try:
                        dir_size += os.path.getsize(filename)
                    except:
                        pass
    return dir_size


#-------------------------------------------------------------------------------


def WriteBinaryFile(filename, data):
    return local_fs.WriteBinaryFile(filename=filename, data=data)


def ReadBinaryFile(filename, decode_encoding=None):
    return local_fs.ReadBinaryFile(filename=filename, decode_encoding=decode_encoding)


def WriteTextFile(filepath, data):
    return local_fs.WriteTextFile(filepath=filepath, data=data)


def ReadTextFile(filename):
    return local_fs.ReadTextFile(filename=filename)


#-------------------------------------------------------------------------------


def _pack_list(lst):
    """
    The core method, convert list of strings to one big string.
    Every line in the string will store a single item from list. First
    line will keep a number of items. So items in the list should be a
    strings and not contain "\n".\ This is useful to store a list of
    users IDs in the local file.
    """
    return strng.to_text(len(lst)) + u'\n' + u'\n'.join(lst)


def _unpack_list(src):
    """
    The core method, read a list from string.
    Return a tuple : (resulted list, list with lines from rest string or None).
    First line of the ``src`` should contain a number of items in the list.
    """
    if not src.strip():
        return list(), None
    words = src.splitlines()
    if len(words) == 0:
        return list(), None
    try:
        length = int(words[0])
    except:
        return words, None
    res = words[1:]
    if len(res) < length:
        res += [
            u'',
        ]*(length - len(res))
    elif len(res) > length:
        return res[:length], res[length:]
    return res, None


def _read_list(path):
    """
    Read list from file on disk.
    """
    src = ReadTextFile(path)
    if src is None:
        return None
    return _unpack_list(src)[0]


def _write_list(path, lst):
    """
    Write a list to the local file.
    """
    return WriteTextFile(path, _pack_list(lst))


def _pack_dict(dictionary, sort=False):
    """
    The core method, convert dictionary to the string.

    Every line in resulted string will contain a key, value pair,
    separated with single space. So keys and must not contain spaces.
    Values must not contain new lines. If ``sort`` is True the resulted
    string will be sorted by keys.
    """
    if sort:
        seq = sorted(dictionary.keys())
    else:
        seq = list(dictionary.keys())
    return u'\n'.join([u'%s %s' % (k, strng.to_text(strng.to_text(dictionary[k]))) for k in seq])


def _unpack_dict_from_list(lines):
    """
    Read dictionary from list, every item in the list is a string with (key,
    value) pair, separated with space.
    """
    dct = {}
    for line in lines:
        words = line.split(u' ')
        if len(words) < 2:
            continue
        dct[words[0]] = u' '.join(words[1:])
    return dct


def _unpack_dict(src):
    """
    The core method, creates dictionary from string.
    """
    lines = strng.to_text(src).split(u'\n')
    return _unpack_dict_from_list(lines)


def _read_dict(path, default=None):
    """
    Read dictionary from local file.

    If file not exist or no read access - returns ``default`` value.
    """
    src = ReadTextFile(path)
    if src is None:
        return default
    return _unpack_dict(src.strip())


def _write_dict(path, dictionary, sort=False):
    """
    Write dictionary to the file.
    """
    data = _pack_dict(dictionary, sort)
    return WriteTextFile(path, data)


def _dir_exist(path):
    """
    Just calls os.path.isdir() method.
    """
    return os.path.isdir(path)


def _dir_make(path):
    """
    Creates a new folder on disk, call built-in os.mkdir() method.

    Set access mode to 0777.
    """
    os.mkdir(path, 0o777)


def _dirs_make(path):
    """
    Create a new folder and all sub dirs, call built-in os.makedirs() method.

    Set access mode to 0777.
    """
    os.makedirs(path, 0o777)


def _dir_remove(path):
    """
    Remove directory recursively.
    """
    rmdir_recursive(path)


#------------------------------------------------------------------------------


def backup_and_remove(path):
    """
    Backup and remove the file.

    Backed up file will have ".backup" at the end. In fact it just tries
    to rename the original file, but also do some checking before. If
    file with ".backup" already exist it will try to remove it before.
    """
    bkpath = path + '.backup'
    if not os.path.exists(path):
        return
    if os.path.exists(bkpath):
        try:
            os.remove(bkpath)
        except:
            lg.out(1, 'bpio.backup_and_remove ERROR can not remove file ' + bkpath)
            lg.exc()
    try:
        os.rename(path, bkpath)
    except:
        lg.out(1, 'bpio.backup_and_remove ERROR can not rename file %s to %s' % (path, bkpath))
        lg.exc()
    if os.path.exists(path):
        try:
            os.remove(path)
        except:
            lg.out(1, 'bpio.backup_and_remove ERROR can not remove file ' + path)
            lg.exc()


def restore_and_remove(path, overwrite_existing=False):
    """
    Restore file and remove the backed up copy.

    Just renames the file with ".backup" at the end to its 'true' name.
    This is reverse method to ``backup_and_remove``.
    """
    bkpath = path + '.backup'
    if not os.path.exists(bkpath):
        return
    if os.path.exists(path):
        if not overwrite_existing:
            return
        try:
            os.remove(path)
        except:
            lg.out(1, 'bpio.restore_and_remove ERROR can not remove file ' + path)
            lg.exc()
    try:
        os.rename(bkpath, path)
    except:
        lg.out(1, 'bpio.restore_and_remove ERROR can not rename file %s to %s' % (path, bkpath))
        lg.exc()


def remove_backuped_file(path):
    """
    Tries to remove the file with ".backup" at the end.
    """
    bkpath = path + '.backup'
    if not os.path.exists(bkpath):
        return
    try:
        os.remove(bkpath)
    except:
        lg.out(1, 'bpio.remove_backuped_file ERROR can not remove file ' + bkpath)
        lg.exc()


#------------------------------------------------------------------------------


def LowerPriority():
    """
    Platform dependent method to lower the priority of the running process.
    """
    try:
        sys.getwindowsversion()  # @UndefinedVariable
    except:
        isWindows = False
    else:
        isWindows = True
    if isWindows:
        # Based on:
        #   "Recipe 496767: Set Process Priority In Windows" on ActiveState
        #   http://code.activestate.com/recipes/496767/
        import win32api  # @UnresolvedImport
        import win32process  # @UnresolvedImport
        import win32con  # @UnresolvedImport
        pid = win32api.GetCurrentProcessId()
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetPriorityClass(handle, win32process.BELOW_NORMAL_PRIORITY_CLASS)
    else:
        import os
        os.nice(20)


def HigherPriority():
    try:
        sys.getwindowsversion()  # @UndefinedVariable
    except:
        isWindows = False
    else:
        isWindows = True
    if isWindows:
        import win32api  # @UnresolvedImport
        import win32process  # @UnresolvedImport
        import win32con  # @UnresolvedImport
        pid = win32api.GetCurrentProcessId()
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetPriorityClass(handle, win32process.REALTIME_PRIORITY_CLASS)
    else:
        import os
        os.nice(1)


#-------------------------------------------------------------------------------


def shortPath(path):
    """
    Get absolute 'short' path in Unicode, converts to 8.3 windows filenames.
    """
    path_ = os.path.abspath(path)
    if not Windows():
        return strng.to_text(path_)
    if not os.path.exists(path_):
        if os.path.isdir(os.path.dirname(path_)):
            res = shortPath(os.path.dirname(path_))
            return strng.to_text(os.path.join(res, os.path.basename(path_)))
        return strng.to_text(path_)
    try:
        import win32api  # @UnresolvedImport
        spath = win32api.GetShortPathName(path_)
        return strng.to_text(spath)
    except:
        lg.exc()
        return strng.to_text(path_)


def longPath(path):
    """
    Get absolute 'long' path in Unicode, convert to full path, even if it was
    in 8.3 format.
    """
    path_ = os.path.abspath(path)
    if not Windows():
        return strng.to_text(path_)
    if not os.path.exists(path_):
        return strng.to_text(path_)
    try:
        import win32api  # @UnresolvedImport
        lpath = win32api.GetLongPathName(path_)
        return strng.to_text(lpath)
    except:
        lg.exc()
    return strng.to_text(path_)


# def portablePath(path):
#    """
#    For Windows changes all separators to Linux format:
#        - "\\" -> "/"
#        - "\" -> "/"
#    If ``path`` is unicode convert to utf-8.
#    """
#    p = path
#    if Windows():
#        p = p.replace('\\\\', '/').replace('\\', '/')
#    if isinstance(p, unicode):
#        return p.encode('utf-8')
#    return p


def remotePath(path):
    """
    Simplify and clean "remote" path value.
    """
    p = strng.to_text(path)
    if p == u'' or p == u'/':
        return p
    p = p.lstrip(u'/').lstrip(u'\\')
    if p.endswith(u'/') and len(p) > 1:
        p = p.rstrip(u'/')
    return p


def portablePath(path):
    """
    Fix path to fit for our use:

    - do convert to absolute path
    - for Windows:
        - change all separators to Linux format: "\\"->"/" and "\"=>"/"
        - convert disk letter to lower case
    - convert to unicode
    """
    path = strng.to_text(path)
    if path == u'' or path == u'/':
        return path
    if Windows() and len(path) == 2 and path[1] == u':':
        # "C:" -> "C:/"
        path += u'/'
    if path.count(u'~'):
        path = os.path.expanduser(path)
    try:
        p = os.path.abspath(path)
    except:
        lg.exc()
        p = path
    if Windows():
        p = p.replace(u'\\', u'/')
        if len(p) >= 2:
            if p[1] == u':':
                p = p[0].lower() + p[1:]
            elif p[:2] == u'//':
                p = u'\\\\' + p[2:]
    if p.endswith(u'/') and len(p) > 1:
        p = p.rstrip(u'/')
    return p


def pathExist(localpath):
    """
    My own "portable" version of built-in ``os.path.exist()`` method.
    """
    if os.path.exists(localpath):
        return True
    p = portablePath(localpath)
    if os.path.exists(p):
        return True
    if Windows() and pathIsNetworkLocation(localpath):
        return True
    return False


def pathIsDir(localpath):
    """
    Assume localpath is exist and return True if this is a folder.
    """
    if os.path.isdir(localpath):
        return True
    if os.path.exists(localpath) and os.path.isfile(localpath):
        return False
    # don't know... let's try portable path
    p = portablePath(localpath)
    if os.path.isdir(p):
        return True
    if os.path.exists(localpath) and os.path.isfile(p):
        return False
    if Windows() and pathIsNetworkLocation(localpath):
        return True
    # may be path is not exist at all?
    if not os.path.exists(localpath):
        return False
    if not os.path.exists(p):
        return False
    # ok, on Linux we have devices, mounts, links ...
    if Linux():
        try:
            import stat
            st = os.path.stat(localpath)  # @UndefinedVariable
            return stat.S_ISDIR(st.st_mode)
        except:
            return False
    # now we are in really big trouble
    raise Exception('path is not exist: %s' % p)
    return False


def pathIsDriveLetter(path):
    """
    Return True if ``path`` is a Windows drive letter.
    """
    p = path.rstrip('/').rstrip('\\')
    if len(p) != 2:
        return False
    if p[1] != ':':
        return False
    if not p[0].isalpha():
        return False
    return True


def pathIsNetworkLocation(path):
    """
    Return True if ``path`` is a Windows network location.

        >>> pathIsNetworkLocation(r'\\remote_machine')
        True
    """
    p = path.rstrip('/').rstrip('\\')
    if len(p) < 3:
        return False
    if not p.startswith('\\\\'):
        return False
    if p[2:].count('\\') or p[2:].count('/'):
        return False
    return True


#------------------------------------------------------------------------------


def main_is_frozen():
    """
    Return True if BitDust is started from .exe not from sources.

    http://www.py2exe.org/index.cgi/HowToDetermineIfRunningFromExe
    """
    return (hasattr(sys, 'frozen') or  # new py2exe
            hasattr(sys, 'importers') or  # old py2exe
            imp.is_frozen('__main__'))  # tools/freeze


def isGUIpossible():
    # if Windows():
    #     return True
    # if Mac():
    #     return True
    # if Linux():
    #     return X11_is_running()
    # All the UI now will be created using Electron framework.
    # To make it possible we need to run local API rest_http/json_rpc server.
    # So here we always return False from now.
    return False


def X11_is_running():
    """
    Linux method to check if BitDust GUI is possible.

    http://stackoverflow.com/questions/1027894/detect-if-x11-is-
    available-python
    """
    global X11isRunning
    if not Linux():
        return False
    if X11isRunning is not None:
        return X11isRunning
    try:
        from subprocess import Popen, PIPE
        p = Popen(['xset', '-q'], stdout=PIPE, stderr=PIPE)
        p.communicate()
        result = p.returncode == 0
    except:
        result = False
    X11isRunning = result
    return X11isRunning


#------------------------------------------------------------------------------


def getExecutableDir():
    """
    A smart way to detect the path of executable folder.
    """
    if main_is_frozen():
        path = os.path.dirname(os.path.abspath(sys.executable))
    else:
        try:
            path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        except:
            path = os.path.dirname(os.path.abspath(sys.argv[0]))
    return strng.to_text(path)


def getExecutableFilename():
    """
    A smart way to detect executable file name.
    """
    if main_is_frozen():
        path = os.path.abspath(sys.executable)
    else:
        path = os.path.abspath(sys.argv[0])
    return strng.to_text(path)


def getUserName():
    """
    Return current user name in unicode string.
    """
    try:
        import pwd
    except ImportError:
        try:
            import getpass
        except:
            pass
        pwd = None
    try:
        if pwd:
            return pwd.getpwuid(os.geteuid()).pw_name
        else:
            return getpass.getuser()
    except:
        pass
    return os.path.basename(strng.to_text(os.path.expanduser('~')))


#------------------------------------------------------------------------------


def listHomeDirLinux():
    """
    Just return a list of files and folders in the user home dir.
    """
    if Windows():
        return []
    rootlist = []
    homedir = os.path.expanduser('~')
    for dirname in os.listdir(homedir):
        if os.path.isdir(os.path.join(homedir, dirname)):
            rootlist.append(dirname)
    return rootlist


def listLocalDrivesWindows():
    """
    Return a list of drive letters under Windows.

    This list should include only "fixed", "writable" and "real" drives,
    not include cd drives, network drives, USB drives, etc.
    """
    if not Windows():
        return []
    rootlist = []
    try:
        import win32api  # @UnresolvedImport
        import win32file  # @UnresolvedImport
        drives = (drive for drive in win32api.GetLogicalDriveStrings().split('\000') if drive)
        for drive in drives:
            if win32file.GetDriveType(drive) == 3:
                rootlist.append(drive)
    except:
        lg.exc()
    return rootlist


def listRemovableDrivesWindows():
    """
    Return a list of "removable" drives under Windows.
    """
    l = []
    try:
        import win32file  # @UnresolvedImport
        drivebits = win32file.GetLogicalDrives()
        for d in range(1, 26):
            mask = 1 << d
            if drivebits & mask:
                # here if the drive is at least there
                drname = '%c:\\' % chr(ord('A') + d)
                t = win32file.GetDriveType(drname)
                if t == win32file.DRIVE_REMOVABLE:
                    l.append(drname)
    except:
        lg.exc()
    return l


def listRemovableDrivesLinux():
    """
    Return a list of "removable" drives under Linux.

    The same idea with ``listRemovableDrivesWindows``.
    """
    try:
        return [os.path.join('/media', x) for x in os.listdir('/media')]
    except:
        return []


def listRemovableDrives():
    """
    Platform-independent way to get a list of "removable" locations.

    Used to detect the location to write a copy of Private Key. Should
    detect USB flash drives.
    """
    if Linux():
        return listRemovableDrivesLinux()
    elif Windows():
        return listRemovableDrivesWindows()
    return []


def listMountPointsLinux():
    """
    Return a list of mount points under Linux.

    Used to detect locations for donated space and local backups.
    """
    mounts = os.popen('mount')
    result = []
    if Linux():
        mch = re.compile('^(.+?) on (.+?) type .+?$')
    else:  # Mac
        mch = re.compile('^(.+?) on (.+?).*?$')
    # mo = re.match('^(.+?) on (.+?) type .+?$', line)
    for line in mounts.readlines():
        mo = mch.match(line)
        if mo:
            device = mo.group(1)
            mount_point = mo.group(2)
            if device.startswith('/dev/'):
                result.append(mount_point)
    return result


def getMountPointLinux(path):
    """
    Return mount point for given path.
    """
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path


#-------------------------------------------------------------------------------


def find_process(applist):
    """
    A portable method to search executed processes.

    You can provide a name or regexp to scan.
    """
    if Android():
        return []
    try:
        import psutil
        pidsL = []
        for p in psutil.process_iter():
            try:
                p_pid = p.pid
                p_cmdline = p.cmdline()
            except:
                continue
            if p_pid == os.getpid():
                continue
            try:
                cmdline = ' '.join(p_cmdline)
            except:
                continue
            for app in applist:
                if app.startswith('regexp:'):
                    if re.match(app[7:], cmdline) is not None:
                        pidsL.append(p_pid)
                else:
                    if cmdline.count(app):
                        pidsL.append(p_pid)
        if pidsL:
            return pidsL
    except:
        pass
    if Windows():
        return find_process_win32(applist)
    else:
        return find_process_linux(applist)
    return []


def kill_process(pid):
    """
    Call to OS ``kill`` procedure. Portable.

    ``pid`` - process id.
    """
    if Android():
        return
    ostype = platform.uname()[0]
    if ostype == 'Windows':
        kill_process_win32(pid)
    else:
        kill_process_linux(pid)


def list_processes_linux():
    """
    This function will return an iterator with the process pid/cmdline tuple.

    :return: pid, cmdline tuple via iterator
    :rtype: iterator

    >>> for procs in list_processes_linux():
    >>>     print procs
    ('5593', '/usr/lib/mozilla/kmozillahelper')
    ('6353', 'pickup -l -t fifo -u')
    ('6640', 'kdeinit4: konsole [kdeinit]')
    ('6643', '/bin/bash')
    ('7451', '/usr/bin/python /usr/bin/ipython')
    """
    for pid_path in glob.glob('/proc/[0-9]*'):
        try:
            # cmdline represents the command whith which the process was started
            f = open('%s/cmdline' % pid_path)
            pid = pid_path.split('/')[2]  # get the PID
            # we replace the \x00 to spaces to make a prettier output from kernel
            cmdline = f.read().replace('\x00', ' ').rstrip()
            f.close()
            yield (pid, cmdline)
        except:
            pass


def find_process_linux(applist):
    """
    You can look for some process name, give a keywords or regexp strings list
    to search.

    This is for Linux.
    """
    pidsL = []
    for pid, cmdline in list_processes_linux():
        try:
            pid = int(pid)
        except:
            continue
        if pid == os.getpid():
            continue
        for app in applist:
            if app.startswith('regexp:'):
                if re.match(app[7:], cmdline) is not None:
                    pidsL.append(pid)
            else:
                if cmdline.find(app) > -1:
                    pidsL.append(pid)
    return pidsL


def find_process_win32(applist):
    """
    Search for process name, for MS Windows.
    """
    pidsL = []
    try:
        import win32com.client  # @UnresolvedImport
        objWMI = win32com.client.GetObject('winmgmts:\\\\.\\root\\CIMV2')
        colProcs = objWMI.ExecQuery('SELECT * FROM Win32_Process')
        for Item in colProcs:
            pid = int(Item.ProcessId)
            if pid == os.getpid():
                continue
            cmdline = Item.Caption.lower()
            if Item.CommandLine:
                cmdline += Item.CommandLine.lower()
            for app in applist:
                if app.startswith('regexp:'):
                    if re.match(app[7:], cmdline) is not None:
                        pidsL.append(pid)
                else:
                    if cmdline.find(app) > -1:
                        pidsL.append(pid)
    except:
        lg.exc()
    return pidsL


def kill_process_linux(pid):
    """
    Make a call to system ``kill`` command.
    """
    try:
        import signal
        os.kill(pid, signal.SIGKILL)
    except:
        lg.exc()


def kill_process_win32(pid):
    """
    Call to system Windows API ``TerminateProcess`` method.
    """
    try:
        from win32api import TerminateProcess, OpenProcess, CloseHandle  # @UnresolvedImport
    except:
        lg.exc()
        return False
    try:
        PROCESS_TERMINATE = 1
        handle = OpenProcess(PROCESS_TERMINATE, False, pid)
    except:
        lg.out(2, 'bpio.kill_process_win32 can not open process %d' % pid)
        return False
    try:
        TerminateProcess(handle, -1)
    except:
        lg.out(2, 'bpio.kill_process_win32 can not terminate process %d' % pid)
        return False
    try:
        CloseHandle(handle)
    except:
        lg.exc()
        return False
    return True


#------------------------------------------------------------------------------


def find_main_process(pid_file_path=None, extra_lookups=[], check_processid_file=True):
    if Android():
        return []
    q = [
        'bitdust-node.exe',
        'bitdust-console.exe',
        'bitdustnode.exe',
        'BitDustNode.exe',
        'BitDustConsole.exe',  # 'bitdust.py',
        'BitDust-node',
        'bpmain.py',
    ]
    if os.environ.get('BITDUST_IN_DOCKER') == '1':
        q.extend([
            'regexp:^.*python.*bitdust.py.*?$',
        ])
    else:
        q.extend([
            'regexp:^.*(?<!\/root\/\.bitdust\/venv\/bin\/)python.*bitdust.py.*?$',
            'regexp:^.*(?<!\/root\/\.bitdust\/venv\/bin\/)Python.*bitdust.py.*?$',
        ])
    q += extra_lookups
    appList = find_process(q)
    if not appList:
        return []
    if not check_processid_file:
        return appList
    try:
        if not pid_file_path:
            from bitdust.main import settings
            pid_file_path = os.path.join(settings.AppDataDir(), 'processid')
        processid = int(ReadTextFile(pid_file_path))
    except:
        processid = None
    if not processid:
        return appList
    if processid not in appList:
        try:
            os.remove(pid_file_path)
        except:
            pass
        return appList
    return [
        processid,
    ]


def lookup_main_process():
    q = [
        'bitdust-node.exe',
        'bitdust-console.exe',
        'bitdustnode.exe',
        'BitDustNode.exe',
        'BitDustConsole.exe',
        'BitDust-node',
        'bpmain.py',
    ]
    if os.environ.get('BITDUST_IN_DOCKER') == '1':
        q.extend([
            'regexp:^.*python.*bitdust.py.*?$',
        ])
    else:
        q.extend([
            'regexp:^.*(?<!\/root\/\.bitdust\/venv\/bin\/)python.*bitdust.py.*?$',
            'regexp:^.*(?<!\/root\/\.bitdust\/venv\/bin\/)Python.*bitdust.py.*?$',
        ])
    return find_process(q)


#------------------------------------------------------------------------------


def detect_number_of_cpu_cores():
    """
    Detects the number of effective CPUs in the system.
    """
    # for Linux, Unix and MacOS
    if hasattr(os, 'sysconf'):
        if 'SC_NPROCESSORS_ONLN' in os.sysconf_names:
            #Linux and Unix
            ncpus = os.sysconf('SC_NPROCESSORS_ONLN')
            if isinstance(ncpus, int) and ncpus > 0:
                return ncpus
        else:
            # MacOS X
            return int(os.popen2('sysctl -n hw.ncpu')[1].read())  # @UndefinedVariable
    # for Windows
    if 'NUMBER_OF_PROCESSORS' in os.environ:
        ncpus = int(os.environ['NUMBER_OF_PROCESSORS'])
        if ncpus > 0:
            return ncpus
    # return the default value
    return 1



#!/usr/bin/env python
# broadcast_listener.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (broadcast_listener.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: broadcast_listener.

.. role:: red

BitDust broadcast_listener() Automat

EVENTS:
    * :red:`broadcaster-connected`
    * :red:`connect`
    * :red:`disconnect`
    * :red:`incoming-message`
    * :red:`init`
    * :red:`lookup-failed`
    * :red:`message-failed`
    * :red:`outbound-message`
    * :red:`shutdown`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import json

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.transport import callback

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

#------------------------------------------------------------------------------

_BroadcastListener = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _BroadcastListener
    if event is None and not args:
        return _BroadcastListener
    if _BroadcastListener is None:
        # set automat name and starting state here
        _BroadcastListener = BroadcastListener('broadcast_listener', 'AT_STARTUP', _DebugLevel, _Debug)
    if event is not None:
        _BroadcastListener.automat(event, *args, **kwargs)
    return _BroadcastListener


#------------------------------------------------------------------------------


class BroadcastListener(automat.Automat):
    """
    This class implements all the functionality of the ``broadcast_listener()``
    state machine.
    """
    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of broadcast_listener() machine.
        """
        self.broadcaster_idurl = None
        self.incoming_broadcast_message_callback = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #--- AT_STARTUP
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'OFFLINE'
                self.doInit(*args, **kwargs)
        #--- BROADCASTER?
        elif self.state == 'BROADCASTER?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnect' or event == 'lookup-failed':
                self.state = 'OFFLINE'
            elif event == 'broadcaster-connected':
                self.state = 'LISTENING'
                self.doSetBroadcaster(*args, **kwargs)
        #--- LISTENING
        elif self.state == 'LISTENING':
            if event == 'disconnect' or event == 'message-failed':
                self.state = 'OFFLINE'
                self.doRemoveBroadcaster(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doRemoveBroadcaster(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'outbound-message':
                self.doSendMessageToBroadcaster(*args, **kwargs)
            elif event == 'incoming-message':
                self.doNotifyInputMessage(*args, **kwargs)
        #--- OFFLINE
        elif self.state == 'OFFLINE':
            if event == 'connect':
                self.state = 'BROADCASTER?'
                self.doStartBroadcasterLookup(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #--- CLOSED
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.incoming_broadcast_message_callback = args[0]
        callback.append_inbox_callback(self._on_inbox_packet)

    def doStartBroadcasterLookup(self, *args, **kwargs):
        """
        Action method.
        """
        from broadcast import broadcasters_finder
        scope = args[0]
        if not scope:
            scope = []
        broadcasters_finder.A('start', (self.automat, {'action': 'listen', 'scopes': json.dumps(scope)}, []))

    def doSetBroadcaster(self, *args, **kwargs):
        """
        Action method.
        """
        self.broadcaster_idurl = args[0]

    def doRemoveBroadcaster(self, *args, **kwargs):
        """
        Action method.
        """
        self.broadcaster_idurl = None

    def doSendMessageToBroadcaster(self, *args, **kwargs):
        """
        Action method.
        """
        from broadcast import broadcast_service
        outpacket = broadcast_service.packet_for_broadcaster(self.broadcaster_idurl, *args, **kwargs)
        p2p_service.SendBroadcastMessage(outpacket)

    def doNotifyInputMessage(self, *args, **kwargs):
        """
        Action method.
        """
        msg, newpacket = args[0]
        if self.incoming_broadcast_message_callback is not None:
            self.incoming_broadcast_message_callback(msg)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.incoming_broadcast_message_callback = None
        callback.remove_inbox_callback(self._on_inbox_packet)
        self.destroy()
        global _BroadcastListener
        del _BroadcastListener
        _BroadcastListener = None

    #-----------------------------------------------------------------------------

    def _on_inbox_packet(self, newpacket, info, status, error_message):
        if newpacket.Command == commands.Broadcast():
            from broadcast import broadcast_service
            msg = broadcast_service.read_message_from_packet(newpacket)
            if not msg:
                lg.warn('not valid message in payload')
                return False
            if newpacket.CreatorID == self.broadcaster_idurl:
                # message from broadcaster - process incoming broadcast
                self.automat('incoming-message', (msg, newpacket))
                return True
            else:
                lg.warn('received broadcast message from another broadcaster? : %s != %s' % (newpacket.CreatorID, self.broadcaster_idurl))
        return False



#!/usr/bin/python
# service_customer_contracts.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_customer_contracts.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_customer_contracts
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return CustomerContractsService()


class CustomerContractsService(LocalService):

    service_name = 'service_customer_contracts'
    config_path = 'services/customer-contracts/enabled'

    def dependent_on(self):
        return [
            'service_customer',
            'service_blockchain_id',
        ]

    def start(self):
        from twisted.internet import task  # @UnresolvedImport
        self.payment_loop = task.LoopingCall(self.on_payment_task)
        self.payment_loop.start(60*60, now=True)
        return True

    def stop(self):
        if self.payment_loop and self.payment_loop.running:
            self.payment_loop.stop()
        self.payment_loop = None
        return True

    def on_payment_task(self):
        from bitdust.customer import payment
        payment.pay_for_storage()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from bitdust_forks.CodernityDB3.misc import NONE


def __patch(obj, name, new):
    n = NONE()
    orig = getattr(obj, name, n)
    if orig is not n:
        if orig == new:
            raise Exception("Shouldn't happen, new and orig are the same")
        setattr(obj, name, new)
    return


def patch_cache_lfu(lock_obj):
    """
    Patnches cache mechanizm to be thread safe (gevent ones also)

    .. note::

       It's internal CodernityDB mechanizm, it will be called when needed

    """
    from . import lfu_cache
    from . import lfu_cache_with_lock
    lfu_lock1lvl = lfu_cache_with_lock.create_cache1lvl(lock_obj)
    lfu_lock2lvl = lfu_cache_with_lock.create_cache2lvl(lock_obj)
    __patch(lfu_cache, 'cache1lvl', lfu_lock1lvl)
    __patch(lfu_cache, 'cache2lvl', lfu_lock2lvl)


def patch_cache_rr(lock_obj):
    """
    Patches cache mechanizm to be thread safe (gevent ones also)

    .. note::

       It's internal CodernityDB mechanizm, it will be called when needed

    """
    from . import rr_cache
    from . import rr_cache_with_lock
    rr_lock1lvl = rr_cache_with_lock.create_cache1lvl(lock_obj)
    rr_lock2lvl = rr_cache_with_lock.create_cache2lvl(lock_obj)
    __patch(rr_cache, 'cache1lvl', rr_lock1lvl)
    __patch(rr_cache, 'cache2lvl', rr_lock2lvl)


def patch_flush_fsync(db_obj):
    """
    Will always execute index.fsync after index.flush.

    .. note::

       It's for advanced users, use when you understand difference between `flush` and `fsync`, and when you definitely need that.

    It's important to call it **AFTER** database has all indexes etc (after db.create or db.open)

    Example usage::

        ...
        db = Database('/tmp/patch_demo')
        db.create()
        patch_flush_fsync(db)
        ...

    """

    def always_fsync(ind_obj):
        def _inner():
            ind_obj.orig_flush()
            ind_obj.fsync()
        return _inner

    for index in db_obj.indexes:
        setattr(index, 'orig_flush', index.flush)
        setattr(index, 'flush', always_fsync(index))

    setattr(db_obj, 'orig_flush', db_obj.flush)
    setattr(db_obj, 'flush', always_fsync(db_obj))

    return



#!/usr/bin/python
# accouning.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (accounting.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: accouning.

Various methods to keep track of:

    + donated space
    + needed space
    + used space
    + free space
    + consumed space
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 4

#------------------------------------------------------------------------------

import os
import math

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import diskusage

from bitdust.lib import misc
from bitdust.lib import strng
from bitdust.lib import jsn
from bitdust.lib import utime

from bitdust.main import settings

from bitdust.contacts import contactsdb

from bitdust.userid import id_url

from bitdust.storage import backup_fs

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'accounting.init')


#------------------------------------------------------------------------------


def read_customers_quotas():
    space_dict = bpio._read_dict(settings.CustomersSpaceFile(), {})
    free_space = int(space_dict.pop('free', 0))
    space_dict = {id_url.field(k).to_bin(): v for k, v in space_dict.items()}
    return space_dict, free_space


def write_customers_quotas(new_space_dict, free_space):
    space_dict = {id_url.field(k).to_text(): v for k, v in new_space_dict.items()}
    space_dict['free'] = free_space
    return bpio._write_dict(settings.CustomersSpaceFile(), space_dict)


def get_customer_quota(customer_idurl):
    customer_idurl = id_url.field(customer_idurl).to_bin()
    try:
        return int(read_customers_quotas()[0].get(customer_idurl, None))
    except:
        return None


def check_create_customers_quotas(donated_bytes=None):
    if not os.path.isfile(settings.CustomersSpaceFile()):
        bpio._write_dict(settings.CustomersSpaceFile(), {
            'free': donated_bytes or settings.getDonatedBytes(),
        })
        lg.info('created a new customers quotas file: %s' % settings.CustomersSpaceFile())
        return True
    return False


def count_consumed_space(space_dict=None):
    if space_dict is None:
        space_dict = read_customers_quotas()[0]
    consumed_bytes = 0
    for idurl, customer_consumed in space_dict.items():
        consumed_bytes += int(customer_consumed)
    return consumed_bytes


def validate_customers_quotas(space_dict=None, free_space=None):
    unknown_customers = set()
    unused_quotas = set()
    if space_dict is None or free_space is None:
        space_dict, free_space = read_customers_quotas()
    for idurl in list(space_dict.keys()):
        idurl = strng.to_bin(idurl)
        try:
            space_dict[idurl] = int(space_dict[idurl])
        except:
            unknown_customers.add(idurl)
            continue
        if space_dict[idurl] <= 0:
            unknown_customers.add(idurl)
            continue
    for idurl in contactsdb.customers():
        if idurl.to_bin() not in list(space_dict.keys()):
            unknown_customers.add(idurl)
    for idurl in space_dict.keys():
        if idurl not in id_url.to_bin_list(contactsdb.customers()):
            unused_quotas.add(idurl)
    return unknown_customers, unused_quotas


#------------------------------------------------------------------------------


def read_customers_usage():
    usage_dict = jsn.dict_keys_to_bin(bpio._read_dict(settings.CustomersUsedSpaceFile(), {}))
    usage_dict = {id_url.field(k).to_bin(): v for k, v in usage_dict.items()}
    return usage_dict


def update_customers_usage(new_space_usage_dict):
    usage_dict = {id_url.field(k).to_bin(): v for k, v in new_space_usage_dict.items()}
    return bpio._write_dict(settings.CustomersUsedSpaceFile(), jsn.dict_keys_to_text(usage_dict))


def calculate_customers_usage_ratio(space_dict=None, used_dict=None):
    if space_dict is None:
        space_dict, _ = read_customers_quotas()
    if used_dict is None:
        used_dict = read_customers_usage()
    current_customers = contactsdb.customers()
    used_space_ratio_dict = {}
    for idurl in current_customers:
        allocated_bytes = int(space_dict[idurl.to_bin()])
        try:
            files_size = int(used_dict.get(idurl.to_bin(), 0))
        except:
            lg.exc()
            files_size = 0
        try:
            ratio = float(files_size)/float(allocated_bytes)
        except:
            lg.exc()
            continue
        used_space_ratio_dict[idurl.to_bin()] = ratio
    return used_space_ratio_dict


#------------------------------------------------------------------------------


def report_consumed_storage():
    result = {}
    result['suppliers_num'] = contactsdb.num_suppliers()
    result['needed'] = settings.getNeededBytes()
    # result['needed_str'] = diskspace.MakeStringFromBytes(result['needed'])
    result['used'] = int(backup_fs.sizebackups())
    # result['used_str'] = diskspace.MakeStringFromBytes(result['used'])
    result['available'] = result['needed'] - result['used']
    # result['available_str'] = diskspace.MakeStringFromBytes(result['available'])
    result['needed_per_supplier'] = 0
    result['used_per_supplier'] = 0
    result['available_per_supplier'] = 0
    if result['suppliers_num'] > 0:
        result['needed_per_supplier'] = int(math.ceil(2.0*result['needed']/result['suppliers_num']))
        result['used_per_supplier'] = int(math.ceil(2.0*result['used']/result['suppliers_num']))
        result['available_per_supplier'] = result['needed_per_supplier'] - result['used_per_supplier']
    # result['needed_per_supplier_str'] = diskspace.MakeStringFromBytes(result['needed_per_supplier'])
    # result['used_per_supplier_str'] = diskspace.MakeStringFromBytes(result['used_per_supplier'])
    # result['available_per_supplier_str'] = diskspace.MakeStringFromBytes(result['available_per_supplier'])
    try:
        result['used_percent'] = misc.value2percent(float(result['used']), float(result['needed']))
    except:
        result['used_percent'] = '0%'
    return result


def report_donated_storage():
    space_dict, free_space = read_customers_quotas()
    used_space_dict = read_customers_usage()
    r = {}
    r['customers_num'] = contactsdb.num_customers()
    r['customers'] = []
    r['old_customers'] = []
    r['errors'] = []
    r['consumed'] = 0
    r['donated'] = settings.getDonatedBytes()
    # r['donated_str'] = diskspace.MakeStringFromBytes(r['donated'])
    r['real'] = bpio.getDirectorySize(settings.getCustomersFilesDir())
    try:
        r['free'] = int(free_space)
    except:
        r['free'] = 0
    used = 0
    for idurl in id_url.to_bin_list(contactsdb.customers()):
        consumed_by_customer = 0
        used_by_customer = 0
        if idurl not in list(space_dict.keys()):
            r['errors'].append('space consumed by customer %r is unknown' % idurl)
        else:
            try:
                consumed_by_customer = int(space_dict.pop(idurl))
                r['consumed'] += consumed_by_customer
            except:
                r['errors'].append('incorrect value of consumed space for customer %r' % idurl)
                continue
        if idurl in list(used_space_dict.keys()):
            try:
                used_by_customer = int(used_space_dict.pop(idurl))
                used += used_by_customer
            except:
                r['errors'].append('incorrect value of used space for customer %r' % idurl)
                continue
        if consumed_by_customer < used_by_customer:
            r['errors'].append('customer %r currently using more space than requested' % idurl)
        c = {}
        c['idurl'] = strng.to_text(idurl)
        c['used'] = used_by_customer
        # c['used_str'] = diskspace.MakeStringFromBytes(c['used'])
        c['consumed'] = consumed_by_customer
        # c['consumed_str'] = diskspace.MakeStringFromBytes(c['consumed'])
        c['real'] = bpio.getDirectorySize(settings.getCustomerFilesDir(idurl))
        # c['real_str'] = diskspace.MakeStringFromBytes(c['real'])
        r['customers'].append(c)
    r['used'] = used
    # r['used_str'] = diskspace.MakeStringFromBytes(r['used'])
    # r['consumed_str'] = diskspace.MakeStringFromBytes(r['consumed'])
    if r['donated'] != r['free'] + r['consumed']:
        r['errors'].append('total consumed %d and known free %d (%d total) bytes not match with donated %d bytes' % (r['consumed'], r['free'], r['consumed'] + r['free'], r['donated']))
    if r['used'] > r['donated']:
        r['errors'].append('total space used by customers exceed the donated limit')
    if len(space_dict) > 0:
        r['errors'].append('found %d incorrect records of consumed space' % len(space_dict))
    if r['real'] != r['used']:
        r['errors'].append('current info needs update, known size is %d bytes but real is %d bytes' % (r['used'], r['real']))
    old_customers_used = 0
    old_customers_real = 0
    for idurl in used_space_dict.keys():
        real = bpio.getDirectorySize(settings.getCustomerFilesDir(idurl))
        try:
            used = int(used_space_dict[idurl])
        except:
            r['errors'].append('incorrect value of used space for customer %r' % idurl)
            continue
        r['old_customers'].append({
            'idurl': strng.to_text(idurl),
            'used': used,  # 'used_str': diskspace.MakeStringFromBytes(used_space_dict[idurl]),
            'real': real,  # 'real_str': diskspace.MakeStringFromBytes(real),
        })
        old_customers_used += used
        old_customers_real += real
    r['old_customers_used'] = old_customers_used
    r['old_customers_real'] = old_customers_real
    try:
        r['used_percent'] = misc.value2percent(float(r['used']), float(r['donated']), 5)
    except:
        r['used_percent'] = ''
    try:
        r['consumed_percent'] = misc.value2percent(float(r['consumed']), float(r['donated']), 5)
    except:
        r['consumed_percent'] = ''
    return r


def report_local_storage():
    # TODO
    # if customers folder placed outside of BaseDir()
    # need to add: total = total + customers
    r = {}
    r['backups'] = bpio.getDirectorySize(settings.getLocalBackupsDir())
    # r['backups_str'] = diskspace.MakeStringFromBytes(r['backups'])
    r['temp'] = bpio.getDirectorySize(settings.getTempDir())
    # r['temp_str'] = diskspace.MakeStringFromBytes(r['temp'])
    r['customers'] = bpio.getDirectorySize(settings.getCustomersFilesDir())
    # r['customers_str'] = diskspace.MakeStringFromBytes(r['customers'])
    r['total'] = bpio.getDirectorySize(settings.AppDataDir())
    # r['total_str'] = diskspace.MakeStringFromBytes(r['total'])
    dataDriveFreeSpace, dataDriveTotalSpace = diskusage.GetDriveSpace(settings.getCustomersFilesDir())
    if dataDriveFreeSpace is None:
        dataDriveFreeSpace = 0
    r['disktotal'] = int(dataDriveTotalSpace)
    # r['disktotal_str'] = diskspace.MakeStringFromBytes(r['disktotal'])
    r['diskfree'] = int(dataDriveFreeSpace)
    # r['diskfree_str'] = diskspace.MakeStringFromBytes(r['diskfree'])
    try:
        r['total_percent'] = misc.value2percent(float(r['total']), float(r['disktotal']), 5)
    except:
        r['total_percent'] = ''
    try:
        r['diskfree_percent'] = misc.value2percent(float(r['diskfree']), float(r['disktotal']), 5)
    except:
        r['diskfree_percent'] = ''
    return r


#------------------------------------------------------------------------------


def verify_storage_contract(json_data):
    try:
        deny = json_data.get('deny')
    except:
        deny = False
    if deny:
        lg.warn(repr(json_data))
        return False
    try:
        utime.unpack_time(json_data['started'])
        utime.unpack_time(json_data['complete_after'])
        utime.unpack_time(json_data['pay_before'])
        int(json_data['value'])
        int(json_data['allocated_bytes'])
        int(json_data['duration_hours'])
        if json_data.get('ecc_position') is not None:
            int(json_data['ecc_position'])
        if json_data.get('ecc_map'):
            str(json_data['ecc_map'])
        float(json_data['raise_factor'])
    except:
        lg.exc(repr(json_data))
        return False
    try:
        gbh = float(json_data['duration_hours'])*(json_data['allocated_bytes']/(1024.0*1024.0*1024.0))
        if json_data['value'] != gbh:
            raise Exception('invalid contract GBH value')
    except:
        lg.exc()
        return False
    return True
